/*
 * LineIterator.java
 *
 * Created on 7 February 2006, 09:57
 */

package blbutil;

/**
 *
 * @author Brian L Browning
 */

import java.io.*;
import java.util.zip.*;
import java.util.NoSuchElementException;

/**
 * An iterator for text files that iterates through the lines of a test file.
 * The <code>next</code> method returns <code>String</code> instances that
 * correspond to lines of a text file.  Files ending in ".gz" are assumed to
 * be GZIP compressed and each line is uncompressed when read.
 *
 * If any IOExceptions are thrown while reading
 * a file, the IOExceptions are trapped, an appropriate error message
 * is written to standard out, and the Java Virtual Machine is
 * terminated.
 */
public class LineIterator implements FileIterator<String> {

    private BufferedReader in = null;
    private String next = null;
    private final File file;

    /**
     * Creates a LineIterator to iterate through
     * the lines of the specified file.  
     *
     * @param filename the name of the text file that will
     *        be read one line at a time.
     */
    public LineIterator(String filename) {
        this(new File(filename));     
    }

    /**
     * Creates a LineIterator to iterate through
     * the lines of the specified file.  Files ending in ".gz"
     * are assumed to be GZIP compressed.
     *
     * @param file text file that will be read one line at a time.
     */
    public LineIterator(File file) {
        this.file = file;
        try {
            // assume .gz extensions are gzip compressed
            if (file.getAbsolutePath().endsWith(".gz")) {
                GZIPInputStream zis = new GZIPInputStream(new FileInputStream(file));
                in = new BufferedReader(new InputStreamReader(zis));
            }
            else {
                in = new BufferedReader(new FileReader(file));
            }
            next = in.readLine();
        }
        catch(FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }
        catch(IOException e) {
            Utilities.exit("Error reading " + file, e);          
        }      
    }
    
    public boolean hasNext() {
        return (next != null);
    }

    public String next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        String current = next;
        try {
            next = in.readLine();
        }
        catch (IOException e) {
            Utilities.exit("Error reading " + file, e);
        }
        return current;
    }

    public void remove() {
        String s = "remove() is not supported by LineIterator";
        throw new UnsupportedOperationException(s);
    }

    public void close() {
        try {
            in.close();
        }
        catch (IOException e) {
            Utilities.exit("Error closing " + file, e);
        }
        next=null;
    }
    
    public String toString() {
        StringBuffer sb = new StringBuffer("[filee=" + file);
        sb.append("\nnext = " + next);
        sb.append("\nBufferedReader = " + in);
        return sb.toString();
    }


}
