package blbutil;
/*
 * FileUtilities.java
 *
 * Created on August 22, 2003, 9:07 AM
 *
 * @author  blb41890
 */

import java.io.*;
import java.util.*;

/**
 * Utility methods.  Contains methods for reading and writing data and
 * other miscellaneous tasks.
 */
public class Utilities {

    private Utilities() {
        // private constructor to prevent instantiation
    }

    /**
     * Prints the specified string, the specified exception, and a 
     * stack trace to standard out and terminates the java interpreter.
     *
     * @s a message to be printed to standard err.
     * @e an exception to be printed to standard err.
     */
    public static void exit(String s, Exception e) {
        e.printStackTrace();
        System.err.println(e);
        System.err.println(s);
        System.err.println("terminating program.");
        System.exit(1);
    }  
    
    /**
     * Copies the specified text file.
     *
     * @param from text file to copy
     * @param to copy of text file to be created.
     *
     * @throws NullPointerException if 
     * <code>from==null || to==null</code>.
     */
    public static void copyFile(String from, String to) {
        LineIterator it = new LineIterator(from);
        PrintWriter out = Utilities.createPrintWriter(to);
        while (it.hasNext()) {
            out.println(it.next());
        }
        out.close();
        it.close();
    }    
    
   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  Any existing file corresponding 
    * with the specified filename will be deleted.  If the file 
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param filename the name of the file to be opened for output. 
    * 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createPrintWriter(String outFile) {
        return createPrintWriter(new File(outFile));
    }
 
    
   /**
    * Returns a <code>java.io.PrintWriter</code> corresponding to
    * the specified filename.  Any existing file corresponding 
    * with the specified filename will be deleted.  If the file 
    * cannot be opened, an error message is printed and the 
    * java interpreter exits.
    *
    * @param file the file to be opened for output. 
    * 
    * @return a <code>java.io.PrintWriter</code> corresponding to
    * the specified file.
    */    
    public static PrintWriter createPrintWriter(File file) {
        PrintWriter out = null;
        try {
            out = new PrintWriter(new BufferedWriter(new FileWriter(file)));
        }
        catch (IOException e) {
            Utilities.exit("Error opening "+ file, e);
        }
        return out;
    }
    
    /**
     * Returns a <code>java.io.DataOutputStream</code> corresponding to
     * the specified file.  Any existing file corresponding to the 
     * <code>File</code> object will be deleted.   If the file cannot be
     * opened, an error message is printed and the java interpreter exits.
     *
     * @param file the file to be opened for output.  
     *
     * @return a <code>java.io.DataOutputStream</code> corresponding to
     * the specified file.
     */
    public static DataOutputStream createDataOutputStream(File file) {
        OutputStream os = null;
        try {
            os = new FileOutputStream(file);
        }
        catch (FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }
        DataOutputStream out = new DataOutputStream(
                new BufferedOutputStream(os));
        return out;
    }
    
    /*
     * Returns a data input stream reading from the specified file.
     * 
     * @param file a binary file.
     *
     * @return a data input stream reading from the specified file.
     */
    public static DataInputStream openDataInputStream(File file) {
        DataInputStream is = null;
        try {
            is = new DataInputStream(new BufferedInputStream(
                    new FileInputStream(file)));
        }
        catch (FileNotFoundException e) {
            Utilities.exit("Error opening " + file, e);
        }    
        return is;
    }  
  
    /*
     * Closes the specified data input stream.
     * 
     * @param is a data input stream.
     */
    public static void closeDataInputStream(DataInputStream is) {
        try {
            is.close();
        }
        catch (IOException e) {
            Utilities.exit("Error closing input stream: " + is, e);
        }    
    }
    
   /**
     * Returns a temporary <code>File</code>.
     *
     * @param prefix the prefix string to be used in generating the file's 
     * name; must be at least three characters long
     *
     * @return a <code>File</code> with an abstract pathname denoting a 
     * newly created empty file.
     */
    public static File tempFile(String prefix) {
       File tempFile = null;
       try {
            tempFile = File.createTempFile(prefix, null);
       } 
       catch (Exception e) {
           System.out.println("Exception thrown by createTempFile: " + e);
           System.exit(1);
       }
       return tempFile;
    }    
    

    /**
     * Returns an <code>int</code> array obtained from parsing
     * the elements of the specified <code>String</code> array.
     * 
     * 
     * 
     * @param a <code>String</code> array whose elements are
     * parsable integers.
     * @return an <code>int</code> array obtained from parsing the 
     * elements of the specified <code>String</code> array.
     * @throws NumberFormatException if the specified array has a 
     * <code>String</code> element that does not contain a parsable
     * integer.
     */
    public static int[] toIntArray(String[] a) {
        int[] intArray = new int[a.length];
        for (int j=0; j<intArray.length; ++j) {
            intArray[j] = Integer.parseInt(a[j]);
        }
        return intArray;    
    }
    
    /**
     * Returns an <code>int</code> array obtained from parsing
     * the specified elements of the specified <code>String</code> 
     * array .
     * 
     * 
     * 
     * @param a <code>String</code> array whose elements are
     * parsable integers.
     * @param start the index of the first array entry to 
     * be parsed.
     * @param end one more than the index of the last array entry
     * to be parsed.
     * @return an <code>int</code> array of length <code>end - start</code>
     * obtained from parsing the specified elements of the specified 
     * <code>String</code> array.
     * @throws NumberFormatException if the specified array has a 
     * <code>String</code> element that does not contain a parsable
     * integer.
     * @throws ArrayIndexOutOfBoundsException if 
     * <code> start < 0 || end &ge; a.length</code>.
     * @throws NegativeArraySizeException if <code>end > start</code>.
     */
    public static int[] toIntArray(String[] a, final int start, final int end) {
        int[] intArray = new int[end - start];
        for (int j=start; j<end; ++j) {
            intArray[j-start] = Integer.parseInt(a[j]);
        }
        return intArray;            
    }

    /**
     * Returns the comand line with double quotes around
     * any arguments containing white-space. New lines characters
     * <tt>\n</tt> are included in the
     * returned <code>String</code> to prevent any line
     * from exceeding the specified maximum length and to prevent a
     * <tt>'-'</tt> argument from ending a line.
     * 
     * 
     * 
     * @param command a <code>String</code> command
     * @param arguments <code>String</code>array of arguments for the
     *         specified command.
     * @param maxLength maximum line length.  Wrap lines
     *        that exceed maxLength;
     * @return the command line with double quotes around
     *         arguments containing white-space.
     */
    public static String getCommandLine(String command, String[] arguments,
                                        int maxLength) {
        final String space = " ";
        final String indent = "    ";
        StringBuffer sb = new StringBuffer(command);
        int length = sb.length();
        String[] args = (String[]) arguments.clone();
        for (int j=0; j<args.length; ++j) {
            String s = args[j];
            if (s.matches("^[^\"].*\\s+.*[^\"]$")) {
                args[j] = "\"" + s + "\"";
            }
        }
        for (int j=0; j<args.length; ++j) {
            int extraChars = 0;
            String nextArg = args[j];
            length += space.length() + nextArg.length();
            if (args[j].trim().charAt(0)=='-' && (j+1)<args.length) {
                extraChars = space.length() + args[j+1].length();
            }
            if (length + extraChars > maxLength) {
                sb.append("\n" + indent);
                length = indent.length() + nextArg.length();
            }
            else {
                sb.append(space);
            }
            sb.append(nextArg);
        }
        sb.append('\n');
        return sb.toString();
    }

    /**
     * Returns a <code>String</code> equal to the contents of the
     * specified text file.  Prints an error message and
     * terminates the java virtual machine if the specified
     * filename is not found or if there is an error reading
     * the file.
     * 
     * 
     * 
     * @param filename a <code>String</code> filename.
     * @param skipLines a nonnegative integer identifying the 
     * number of lines to skip
     * @return a <code>String</code> equal to the contents of the
     *         specified text file.
     * @throws IllegalArgumentException if <code>skipLines < 0</code>.
     */
    public static String fileToString(String filename, int skipLines) {
        if (skipLines < 0) {
            throw new IllegalArgumentException("skipLines < 0: " + skipLines);
        }
        StringBuilder sb = new StringBuilder();
        try {
            BufferedReader in = new BufferedReader(new FileReader(filename));
            String line;
            for (int j=0; j<skipLines; ++j) {
                in.readLine();
            }
            if ((line=in.readLine()) != null) {
                sb.append(line);
            }
            while ((line=in.readLine()) != null) {
                sb.append('\n');
                sb.append(line);
            }
            in.close();
        }
        catch(FileNotFoundException e) {
            System.out.print("Utilities.fileToString: unable to open: ");
            System.out.println(filename + ".  Terminating program");
            System.exit(0);
        }
        catch(IOException e) {
            System.out.println("Error: IOException.  Terminating programing");
            System.exit(0);
        }
        return sb.toString();
    }

   /**
     * Calls the <code>toString()</code> method of the specified object
     * and writes the returned <code>String</code> to the specified
     * file.  Prints an error message and terminates the java
     * virtual machine if there is an error reading the file.
     * 
     * 
     * 
     * @param o an <code>Object</code> whose <code>String</code>
     *        representation is to be written to the specified
     *        file.
     * @param filename a <code>String</code> filename.
     * @throws NullPointerException if <code>o==null || filename==null</code>.
     */
    public static void objectToFile(Object o, String filename) {
        try {
           Writer w = new FileWriter(filename);
           w.write(o.toString());
           w.close();
        }
        catch (IOException e) {
            System.out.println("Error in printing object: " + e);
            System.exit(1);
        }
    }

    /**
     * Returns a map with one (key, value) pair for each string element
     * of the specified array that contains the specified separator.
     * If an array element <code>s</code> of the specified array contains 
     * the specific separator character then the key is
     * <code> s.substring(0, s.indexOf('sep'))</code>
     * and the value is <code>s.substring(s.indexOf(s.indexOf(sep) + 1)</code>.
     * Thus the emptry string ("") is returned if the first occurrence of 
     * the specified separator is the last character of the string.
     * <p/>
     * 
     * If the array element <code>s</code> of the specified array does not 
     * contain the specified separator character then the entry is ignored.
     * 
     * @param args a string array.
     * @param sep a separator character used to create a (key, value) pair 
     * from a string of the form <code>key + sep + value</code>.
     * @return a <code>java.util.Map<String, String></code> a map with one
     *        (key, value) pair for each element of the specified
     *        string array with the form <code>key + sep + value</code>.
     * @throws NullPointerException if <code>args==null</code>.
     * @throws IllegalArgumentException if any two elements of the 
     * specified string array have the same key.
     */
    public static Map<String,String> argsToMap(String[] args, char sep) {
        Map<String,String> argMap = new HashMap<String,String>();
        for (String arg : args) {
            String value = null;
            int index = arg.indexOf(sep);
            if (index != -1) {
                String key = arg.substring(0, index);
                value = arg.substring(index + 1);
                if (argMap.containsKey(key)) {
                    String s = "duplicate arguments: " + key;
                    throw new IllegalArgumentException(s);
                }
                argMap.put(key, value);                
            }
        }
        return argMap;
    }
    
    /**
     * Returns the string entries that do not contain the specified 
     * separator character as a list.  The elements of the returned list
     * will have the same ordering as in the specified string array.
     *
     * @param args a string array.
     * @param sep a separator character.  Elements of the string array
     * that contains the separator character are ignored.
     *
     * @return the string entries that do not contain the specified 
     * separator character as a list.
     *
     * @throws NullPointerException if <code>args==null</code>.
     */
    public static List<String> argsToList(String[] args, char sep) {
        List<String> argList = new ArrayList<String>();
        for (String arg : args) {
            int index = arg.indexOf(sep);
            if (index==-1) {
                argList.add(arg);
            }
        }
        return argList;
    }

    /**
     * Returns a <code>String</code> ruler identifying character
     * position on printed output.
     * 
     * 
     * 
     * @return a <code>String</code> ruler identifying character
     *         position on printed output.
     */
    public static String ruler() {
        StringBuilder sb = new StringBuilder();
        sb.append("0        10        20        30        ");
        sb.append("40        50        60        70        80         \n");
        sb.append("123456789012345678901234567890123456789");
        sb.append("012345678901234567890123456789012345678901234567890\n");
        return sb.toString();
    }

    /**
     * Returns a ragged two-dimensional <code>String</code> array by
     * splitting the specified <code>String</code> on whitespace and
     * new line characters, "\n".  Blank lines are ignored.  The
     * rows and elements of the array correspond to the lines
     * and words of the specified <code>String</code> respectively.
     * 
     * 
     * 
     * @param s a <code>String</code> to be split into a two-dimensional
     *        array.
     * @return a two-dimensional <code>String</code> array corresponding
     *         to the specified <code>String</code> split on whitespace
     *         and new line characters.
     * @throws NullPointerException if <code>s==null</code>.
     */
    public static String[][] stringTo2DArray(String s) {
        String[] lines = s.split("\\s*\n\\s*");
        String[][] values = new String[lines.length][];
        for (int i=0; i<values.length; ++i) {
            values[i] = lines[i].split("\\s+");
        }
        return values;

    }

    /**
     * Removes any <code>String</code> consisting of whitespace
     * from the specified <code>String</code> array.  The order
     * of the array elements is preserved.  A <code>String</code>
     * <tt>s</tt> is removed if <code>!s.trim().equals("")</code>.
     * 
     * 
     * 
     * @param lines a <code>String</code> array.
     * @return an array consisting of all non-blank <code>String</code>
     *         elements of the specified <code>String</code>array.
     *         A <code>String</code> <tt>s</tt> is blank if
     *         <code>!s.trim().equals("")</code>.
     */
    public static String[] removeBlankLines(String[] lines) {
       List<String> l = new ArrayList<String>(lines.length);
       for (String s : lines) {
           if (!s.trim().equals("")) {
               l.add(s);
           }
       }
       return (String[]) l.toArray(new String[0]);
    }

     /**
     * Returns a string representation of the specified elapsed time.
     *
     * @param milliseconds the elapsed time in milliseconds
     *
     * @reutrn a string representation of the specified elapsed time.
     */
    public static String elapsedTime(long milliseconds) {
        long seconds = Math.round(milliseconds /1000.0);
        StringBuilder sb = new StringBuilder();
        if (seconds >= 3600) {
            sb.append(seconds / 3600);
            sb.append(" hours ");
            seconds %= 3600;
 
        }
        if (seconds >= 60) { 
            sb.append(seconds / 60);
            sb.append(" minutes ");
            seconds %= 60;
        }
        sb.append(seconds);
        sb.append(" seconds");
        return sb.toString();
    }   
}

