/*
 * CharStringValidator.java
 *
 * Created on 3 June 2006, 14:35
 *
 */

package blbutil.validator;

/**
 * @author Brian L Browning
 */

import blbutil.*;
import java.util.*;

/**
 * Class <code>CharStringValidator</code> is used to validate a map
 * entry whose value is a string.
 */
public class CharStringValidator implements Validator {
    
    private String key = null;
    private String valueType = null;
    private String defaultValue = null;
    private boolean isRequired;
    private boolean isAdvanced;
    private String err = null;
    private char[] values = null;
    
    /**
     * Creates a <code>Validator</code> whose <code>validate</code> method
     * validates that a map contains an entry with the specified key
     * and whose value must contain a subset of a specified set of characters.
     *
     * @param key the key in a (key, value) pair.
     * @param valueType short description of value in (key, value) pair.
     * @param defaultValue the default value for the (key, value) pair.  If 
     * <code>defaultValue.equals("")</code> then no default value is defined.
     * @param isRequired a boolean that determines whether an entry with
     * the key mapped to a non-empty value is required to be present in a
     * map in order to validate.
     * @param values a character array giving the alphabet of characters
     * that define the valid values that can be associated with the 
     * specified key.  
     *
     * @throws NullPointerException if 
     * <code>key==null || valueType==null || defaultValue==null 
     * || characters==null</code>.
     * @throws IllegalArgumentException if 
     * <code>isRequired && !defaultValue.equals("")</code> or if
     * <code>defaultValue</code> contains a character that is not
     * an element of the <code>characters</code> array.
     */
    public CharStringValidator(String key, String valueType, String defaultValue, 
            boolean isRequired, boolean isAdvanced, char[] characters) {
        if (key==null) {
            throw new NullPointerException("key==null");
        }
        if (valueType==null) {
            throw new NullPointerException("value==null");
        }
        if (defaultValue==null) {
            throw new NullPointerException("err==null");
        }
        if (characters==null) {
            throw new IllegalArgumentException("characters==null");
        }
        this.values = (char[]) characters.clone();
        Arrays.sort(values);
        if (!defaultValue.equals("")) {
            if (isRequired) {
                String msg = "key required but default value = " + defaultValue;
                throw new IllegalArgumentException(msg);
            }
            else {
                for (int j=0, n=defaultValue.length(); j<n; ++j) {
                    char c = defaultValue.charAt(j);
                    if (Arrays.binarySearch(values, c) < 0) {
                        throw new IllegalArgumentException(err(c, defaultValue));
                    }
                }
            }
        }
        this.key = key;
        this.valueType = null;
        this.defaultValue = defaultValue; 
        this.isRequired = isRequired;
        this.isAdvanced = isAdvanced;
        this.err = key + "=" + valueType;
    }
        
    public void validate(Map<String, String> map) {
        boolean keyFound = map.containsKey(key);
        if (!keyFound) {
            if (isRequired) {
                throw new IllegalArgumentException("Missing argument : " + err);
            }
            else {
                map.put(key, defaultValue);
            }
            return;
        }
        else if (values != null) {
            String value = map.get(key);           
            for (int j=0, n=value.length(); j<n; ++j) {
                char c = value.charAt(j);
                if (Arrays.binarySearch(values, c) < 0) {
                    throw new IllegalArgumentException(err(c, value));
                }
            }
        }
    }
    
    private String err(char c, String value) {
        StringBuilder sb = new StringBuilder();
        sb.append("char '" + c + "' in \"");
        sb.append(value + "\" is not");
        for (int j=0; j<values.length-1; ++j) {
            sb.append(" '");
            sb.append(values[j]);
            sb.append("',");
        } 
        sb.append(" or '" + values[values.length-1] + "'.");
        return sb.toString();
    }
    
    public String key() {
        return key;
    }
    
    public String defaultValue() {
        return defaultValue;
    } 
    
    public boolean isRequired() {
        return isRequired;
    } 
    
    public boolean isAdvanced() {
        return isAdvanced;
    }
    
    public String toString() {
        return err;
    }
}
