/*
 * IntValidator.java
 *
 * Created on 1 June 2006, 13:02
 *
 */

package blbutil.validator;

/**
 * @author Brian L Browning
 */

import java.util.Map;

/**
 * Class <code>IntValidator</code> is used to validate a map
 * containing an entry whose value is a parsable <code>int</code>.
 */
public class IntValidator implements Validator{
    
    private String key = null;
    private String valueType = null;
    private String defaultValue = null;
    private boolean isRequired = true;
    private boolean isAdvanced = true;
    private int min = Integer.MIN_VALUE;
    private int max = Integer.MIN_VALUE;
    private String err = null;                
    
    /**
     * Creates a <code>Validator</code> whose <code>validate</code> method
     * validates that a map contains an entry with the specified key
     * and whose value is greater than or equal to the specified minimum
     * and less than or equal to the specified maximum. 
     *
     * @param key the key in a (key, value) pair.
     * @param valueType short description of value in (key, value) pair.
     * @param defaultValue the default value for the (key, value) pair.  If 
     * <code>defaultValue.equals("")</code> then no default value is defined.
     * @param isRequired a boolean that determines whether an entry with
     * the key mapped to a non-empty value is required to be present in a
     * map in order to validate.
     * @param min the minimum valid integer value
     * @param max the maximum valid integer value
     *
     * @throws NullPointerException if 
     * <code>key==null || valueType==null || defaultValue==null</code>.
     * @throws IllegalArgumentException if 
     * <code>min > max || Integer.parseInt(defaultValue) < min
     * || Integer.parseInt(defaultValue) > max</code> or if 
     * <code>isRequired && !defaultValue.equals("")</code>.
     * @throws NumberFormatException if <code>defaultValue</code> is not
     * the empty string ("") and is not a parsable <code>int</code>.
     */
    public IntValidator(String key, String valueType, String defaultValue, 
            boolean isRequired, boolean isAdvanced, int min, int max) {
        if (min > max) {
            throw new IllegalArgumentException("min=" + min + " > max=" + max);
        }
        if (key==null) {
            throw new NullPointerException("key==null");
        }
        if (valueType==null) {
            throw new NullPointerException("value==null");
        }
        if (defaultValue==null) {
            throw new NullPointerException("err==null");
        }
        if (!defaultValue.equals("")) {
            if (isRequired) {
                String msg = "key required but default value = " + defaultValue;
                throw new IllegalArgumentException(msg);
            }
            int x = Integer.parseInt(defaultValue);
            if (x < min) {
                String s = "defaultValue=" + defaultValue + " < min=" + min;
                throw new IllegalArgumentException(s);
            }
            if (x > max) {
                String s = "defaultValue=" + defaultValue + " > max=" + max;
                throw new IllegalArgumentException(s);                   
            }
        }
        this.key = key;
        this.valueType = valueType;
        this.defaultValue = defaultValue; 
        this.isRequired = isRequired;
        this.isAdvanced = isAdvanced;
        this.min = min;
        this.max = max;
        this.err = key + "=" + valueType;  
    }
        
    public void validate(Map<String, String> map) {
        boolean keyFound = map.containsKey(key);
        if (!keyFound) {
            if (isRequired) {
                throw new IllegalArgumentException("Missing argument : " + err);
            }
            else {
                map.put(key, defaultValue);
            }
            return;
        }
        else {
            String value = map.get(key);
            if (value.equals("")) {
                String msg = "Missing " + valueType + " : " + err;
                throw new IllegalArgumentException(msg);
            }
            try {
                int x = Integer.parseInt(value);
                if (x < min) {
                    String msg = "Error in " + err + " argument.  ";
                    msg += value + " is not >= " + min;
                    throw new IllegalArgumentException(msg);
                }
                if (x > max) {
                    String msg = "Error in " + err + " argument.  ";
                    msg += value + " is not <= " + max;
                    throw new IllegalArgumentException(msg);                  
                }
            } 
            catch (NumberFormatException e) {
                String msg = value + " is not a number : " + err;
                throw new IllegalArgumentException(msg);
            }                
        }
    }
    
    public String key() {
        return key;
    }
    
    public String defaultValue() {
        return defaultValue;
    } 
    
    public boolean isRequired() {
        return isRequired;
    }
    
    public boolean isAdvanced() {
        return isAdvanced;
    }
    
    public String toString() {
        return err;
    }
}
