/*
 * StringValidator.java
 *
 * Created on 1 June 2006, 13:07
 *
 */

package blbutil.validator;

/**
 * @author Brian L Browning
 */
import java.util.*;

/**
 * Class <code>StringValidator</code> is used to validate a map
 * entry whose value is a string.
 */
public class StringValidator implements Validator {
    
    private String key = null;
    private String valueType = null;
    private String defaultValue = null;
    private boolean isRequired = true;
    private boolean isAdvanced = true;
    private String err = null;
    private Set<String> valueSet = null;
    
    /**
     * Creates a <code>Validator</code> whose <code>validate</code> method
     * validates that a map contains an entry with the specified key
     * and whose value is equal to one of the specified valueSet.
     * 
     * @param key the key in a (key, value) pair.
     * @param valueType short description of value in (key, value) pair.
     * @param defaultValue the default value for the (key, value) pair.  If 
     * <code>defaultValue.equals("")</code> then no default value is defined.
     * @param isRequired a boolean that determines whether an entry with
     * the key mapped to a non-empty value is required to be present in a
     * map in order to validate.
     * @param valueSet a set of strings that gives the valid valueSet that
     * can be associated with the specified key.  If <code>valueSet==null</code> 
     * then there are no restrictions on the string that is associated 
     * with the specified key.
     * @throws NullPointerException if 
     * <code>key==null || valueType==null || defaultValue==null </code>.
     * @throws IllegalArgumentException if 
     * <code>isRequired && !defaultValue.equals("")</code>, if
     * <code>valueSet.size()==0</code>, or if 
     * <code>valueSet!=null && !valueSet.contains(defaultValue)</code>.
     */
    public StringValidator(String key, String valueType, String defaultValue, 
            boolean isRequired, boolean isAdvanced, String[] values) {
        if (key==null) {
            throw new NullPointerException("key==null");
        }
        if (valueType==null) {
            throw new NullPointerException("value==null");
        }
        if (defaultValue==null) {
            throw new NullPointerException("defaultValue==null");
        }
        if (values!=null) {
            if (values.length==0) {
                throw new IllegalArgumentException("values.size()==0");
            }
            else {
                this.valueSet = new HashSet<String>(values.length);
                for (String s : values) {
                    this.valueSet.add(s);
                }
            }
        }        
        if (!defaultValue.equals("")) {
            if (isRequired) {
                String msg = "key required but default value = " + defaultValue;
                throw new IllegalArgumentException(msg);
            }
            if (this.valueSet!=null && !this.valueSet.contains(defaultValue)) {
                String s = "defaultValue is not in values: " + defaultValue;
                s += " not in " + values;
                throw new IllegalArgumentException(s);
            }
        }
        this.key = key;
        this.valueType = valueType;
        this.defaultValue = defaultValue;
        this.isRequired = isRequired;
        this.isAdvanced = isAdvanced;
        this.err = key + "=" + valueType;
    }
        
    public void validate(Map<String, String> map) {
        boolean keyFound = map.containsKey(key);
        if (!keyFound) {
            if (isRequired) {
                throw new IllegalArgumentException("Missing argument: " + err);
            }
            else {
                map.put(key, defaultValue);
            }
            return;
        }
        else if (valueSet != null) {
            String value = map.get(key);           
            if (!valueSet.contains(value)) {
                String s = "Error in " + err + " argument.  ";
                s +=  value + " is not in the set: " + valueSet;
                throw new IllegalArgumentException(s);
            }
        }
    }
    
    public String key() {
        return key;
    }
    
    public String defaultValue() {
        return defaultValue;
    } 
    
    public boolean isRequired() {
        return isRequired;
    }
    
    public boolean isAdvanced() {
        return isAdvanced;
    }
    
    public String toString() {
        return err;
    }
}
