/*
 * AbstractMultiFileIterator.java
 *
 * Created on 7 May 2006, 14:49
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import blbutil.FileIterator;
import java.util.NoSuchElementException;

/**
 * Abstract class for iterating through a file obtained by 
 * concatenating a list of files.
 */
public abstract class AbstractMultiFileIterator<E> implements FileIterator<E> {
    
    private FileIterator<E> it = null;
    private E next = null;
    private String[] filenames = null;
    private int index = -Integer.MAX_VALUE;   
    
    /**
     * Creates a new <code>AbstractMultiFileIterator</code>.
     *
     * @param fileNames a list of file names.  The <code>next</code>
     * method will iterate through each file in the specified list.
     *
     * @throws NullPointerException if <code>fileNames==null</code>.
     */
    public AbstractMultiFileIterator(String[] fileNames) {
        this.filenames = (String[]) fileNames.clone();
        if (filenames.length > 0) {
            this.it = createIterator(filenames[0]);
            this.index = 0;
            this.next = readNext();
        }
    }
    
    /*
     * Reads the next value in the file obtained by concatenating
     * a list of files.
     */
    private E readNext() {
        while (!it.hasNext() && ++index < filenames.length) {
            it.close();
            it = createIterator(filenames[index]);
        } 
        return (it.hasNext() ? it.next() : null);
    }
    
    /**
     * Returns a <code>FileIterator<E></code> that iterates
     * through the specified filename.
     * 
     * @param filename name of file.
     * @return a <code>FileIterator<E></code> that iterates
     * through file with the specified filename.
     */
    protected abstract FileIterator<E> createIterator(String filename);
    
    public boolean hasNext() {
         return (next != null);       
    }
    
    public E next() {
        if (next == null) {
            throw new NoSuchElementException();
        }
        E current = next;
        next = readNext();
        return current;
    }

    public void remove() {
        throw new UnsupportedOperationException();
    }
    
    public void close() {
        if (it != null) {
            it.close();
        }
        next=null;
    } 
}
