/*
 * AlleleIndices.java
 *
 * Created on 22 July 2006, 06:45
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;


/*
 * Class <code>AlleleIndices</code> is a package private class used by the 
 * implementation of AltIntDiallelic. 
 */
class AlleleIndices {
    /* 
     * This implementation uses sorted int arrays to store allele indices.
     */
    private final int refAllele;
    private boolean refIsMinor = true;
    private final int alleleCount;
    private final int[] strata;
    private int[] strataCounts;
    private int[][] minorAlleles;
    private int[][] dom;
    private int[][] rec;
    private int[][] missingAlleles;
    private int[][] missingGenotypes;
    
    // fields used during object construction.
    private final boolean isAllelicData;
    private int[] minorCount;    
    private int[] rCount;
    private int[] dCount;
    private int[] missingACount;
    private int[] missingGCount;

    
    /**
     * Creates a new instance of <code>AlleleIndices</code> which stores 
     * for each population strata indices of the minor alleles, 
     * missing alleles, recessive genotypes,
     * dominant genotypes, and missing genotypes for the diallelic marker 
     * determine by the specified reference allele of the specified marker 
     * data.  If the number of alleles is odd, the data is assumed to 
     * represent allelic data and have no recessive, dominant, 
     * or missing genotypes.
     * 
     * @param markerLine genetic marker data. 
     * @param refAllele the integer coding for the reference allele. 
     * @param strata an array of non-negative integers whose k-th element
     * is the population stratum for the k-th allele.
     *
     * @throws IllegalArgumentException if 
     * <code>markerLine.hasMissingCode() && (refAllele==0) </code>, if 
     * <code>strata.length != markerLine.data().length</code>, or if 
     * any element of the specified strata array is negative.
     */    
    public AlleleIndices(MarkerLineInterface markerLine, int refAllele, int[] strata) {
        if ((refAllele==0) && markerLine.hasMissingCode()) {
            throw new IllegalArgumentException("ref allele is missing allele");
        }
        int[] data = markerLine.data();
        if (strata.length != data.length) {
            throw new IllegalArgumentException("strata.length=" + strata.length 
                    + " != data.length=" + data.length);
        }
        boolean hasMissingCode = markerLine.hasMissingCode();
        this.isAllelicData = isAllelicData(strata);
        this.refAllele = refAllele;
        this.alleleCount = data.length;
        this.strata = (int[]) strata.clone();

        if (isAllelicData) {
            countAlleleData(refAllele, data, strata, hasMissingCode);
            setAlleleData(refAllele, data, strata, hasMissingCode);
        }
        else {
            countGenotypeData(refAllele, data, strata, hasMissingCode);
            setGenotypeData(refAllele, data, strata, hasMissingCode);
        }
    }
    
    /*
     * Returns <code>true</code> if the specified array has odd length
     * or if the two array entries for any individual are not identical 
     * and <code>false</code> otherwise.
     *
     * @return <code>true</code> if the specified array has odd length
     * or if the two array entries for any individual are not identical
     * and <code>false</code> otherwise.
     */
    private boolean isAllelicData(int[] array) {
        if ((array.length & 1)==1) {
            return true;
        }
        for (int j=1; j<array.length; j+=2) {
            if (array[j] != array[j-1]) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Returns one plus the largest entry in the specified 
     * <code>int[]</code> array or <code>Integer.MIN_VALUE</code>
     * if the specified array has length 0.
     *
     * @return one plus the largest entry in the specified 
     * <code>int[]</code> array or <code>Integer.MIN_VALUE</code>
     * if the specified array has length 0.
     *
     * @throws NullPointerException if <code>array==null</code>.
     * @throws IllegalArgumentException if any element of the 
     * specified array is negative.
     */
    private int countStrata(int[] array) {
        int max = Integer.MIN_VALUE;
        for (int i : array) {
            if (i > max) {
                max = i;
            }
            if (i < 0) {
                throw new IllegalArgumentException("negative stratum: " + i);
            }
        }
        return (max + 1);
    }
    
    /*
     * Returns the sum of the elements in the specified 
     * <code>int[]</code> array.  Note: the sum can 
     * "wrap-around", and will always lie between 
     * <code>Integer.MIN_VALUE</code> 
     * and <code>Integer.MAX_VALUE</code>.
     *
     * @return the sum of the integers in the specified 
     * array.
     */
    private int sum(int[] array) {
        int sum = 0;
        for (int i : array) {
            sum += i;
        }
        return sum;
    }

    /*
     * Counts minor alleles and missing alleles in each strata, stores
     * these counts in the private <code>this.minorCount</code> and 
     * <code>this.missingACount</code> arrays, allocates
     * the private <code>this.minorAlleles[][]</code> and 
     * <code>this.missingAlleles[][]</code> arrays to store the minor and 
     * missing allele indices.  Sets <code>this.refIsMinor</code> to 
     * <code>false</code> if the reference allele is not the minor allele.
     * If the reference allele is 0 and <code>hasMissing==true</code> 
     * the contract for the <code>countAlleleData</code> method is undefined.
     *
     * @param refAllele the reference allele
     * @param data integer array of allele data
     * @param strata an array of non-negative integers whose k-th element
     * is the population stratum for the k-th allele.
     * @param hasMissing <code>true</code> if denotes a missing allele</code>
     * and <code>false</code> if there is no missing allele code.
     */
    private void countAlleleData(int refAllele, int[] data, int[] strata, 
            boolean hasMissing) {
        int nStrata = countStrata(strata);
        this.strataCounts = new int[nStrata];
        allocateCountArrays(nStrata);
        for (int j=0; j<data.length; ++j) {
            int a = data[j];
            int stratum = strata[j];
            ++strataCounts[stratum];
            if (hasMissing && a==0) {
                ++missingACount[stratum];
            }  
            else if (a==refAllele) {
                ++minorCount[stratum];
            }    
        }
        int totalMinorCount = sum(minorCount);
        int totalMissingACount = sum(missingACount);
        if (totalMinorCount > data.length - totalMissingACount - totalMinorCount) {
            this.refIsMinor = false;
            for (int j=0; j<minorCount.length; ++j) {
                minorCount[j] = strataCounts[j] - missingACount[j] - minorCount[j];
            }
        }
        allocateArrays(strataCounts.length);
    }
    
    /*
     * Stores indices of minor alleles and missing alleles in the
     * private <code>minorAlleles[][]</code> and 
     * <code>missingAlleles[][]</code> arrays.
     *
     * @param refAllele the reference allele
     * @param data integer array of allele data
     * @param hasMissing <code>true</code> if 0 denotes a missing allele</code>
     * and <code>false</code> if there is no missing allele code.
     */    
    private void setAlleleData(int refAllele, int[] data, int[] strata,
            boolean hasMissing) {
        resetCounts();
        for (int j=0; j<data.length; ++j) {
            int a = data[j];
            int stratum = strata[j];
            if (a==refAllele ^ !this.refIsMinor) {
                minorAlleles[stratum][minorCount[stratum]++] = j;
            }
            else if (hasMissing && a==0) {
                missingAlleles[stratum][missingACount[stratum]++] = j;
            }      
        }
    }   

    /*
     * Counts minor alleles, missing alleles, recessive genotypes,
     * dominant genotypes, and missing genotypes in each strata,
     * stores these counts in the private <code>this.minorCount</code>,
     * <code>this.missingACount</code>, <code>this.rCount</code>, 
     * <code>this.dCount</code>, and <code>this.missingGCount</code>,
     * arrays, and allocates the private <code>minorAlleles[][]</code>, 
     * <code>missingAlleles[][]</code>, <code>this.rec[][]</code>,
     * <code>this.dom[][]</code>, and <code>this.missingGenotypes[][]</code>
     * arrays to store allele and genotype indices of this data. 
     * Sets <code>this.refIsMinor</code> to <code>false</code> if the 
     * reference allele is not the minor allele.  
     * If the number of alleles is odd, the last allele is ignored.  
     * If the reference allele is 0 and <code>hasMissing==true</code> 
     * the contract for the <code>countGenotypeData</code> method is undefined.
     *
     * @param refAllele the reference allele
     * @param data integer array of allele data
     * @param strata an array of non-negative integers whose k-th element
     * is the population stratum for the k-th allele.
     * @param hasMissing <code>true</code> if denotes a missing allele</code>
     * and <code>false</code> if there is no missing allele code.
     */   
    private void countGenotypeData(int refAllele, int[] data, int[] strata, 
            boolean hasMissing) {
        int nStrata = countStrata(strata);
        this.strataCounts = new int[nStrata];
        allocateCountArrays(nStrata);
        for (int j=1; j<data.length; j+=2) {
            int a1 = data[j-1];
            int a2 = data[j];
            boolean allele1 = (a1==refAllele);
            boolean allele2 = (a2==refAllele);
            int stratum = strata[j];
            strataCounts[stratum] += 2;
            if (hasMissing && (a1==0 || a2==0))  {
                ++missingGCount[stratum];
                if (a1==0) {
                    ++missingACount[stratum];
                }
                if (a2==0) {
                    ++missingACount[stratum];
                }
                if (allele1 || allele2) {
                    ++minorCount[stratum];
                } 
            }
            else if (allele1 && allele2) {
                minorCount[stratum] += 2;
                ++rCount[stratum];
                ++dCount[stratum];
            }
            else if (allele1 ^ allele2) {
                ++minorCount[stratum];
                ++dCount[stratum];
            }
        }
        int totalMinorCount = sum(minorCount);
        int totalMissingACount = sum(missingACount);
        if (totalMinorCount > data.length - totalMissingACount - totalMinorCount) {
            this.refIsMinor = false;
            for (int j=0; j<strataCounts.length; ++j) {
                minorCount[j] = strataCounts[j] - missingACount[j] - minorCount[j];
                int gCount = strataCounts[j]/ 2 - missingGCount[j];
                int tempCount = dCount[j];
                dCount[j] = gCount - rCount[j];
                rCount[j] = gCount - tempCount;
            }
        } 
        allocateArrays(strataCounts.length);
    }
             
    /*
     * Stores indices of minor alleles, missing alleles, recessive genotypes,
     * dominant genotypes, and missing genotypes, in
     * the private <code>this.minorAlleles[][]</code>, 
     * <code>missingAlleles[][]</code>, <code>rec[][]</code>, 
     * <code>dom[][]</code>, and <code>missingGenotypes[][]</code>
     * arrays.  If the number of alleles is odd, the last allele is ignored.  
     * If the reference allele is 0 and <code>hasMissing==true</code> the 
     * contract for the <code>setGenotypeData</code> method is undefined.
     *
     * @param refAllele the reference allele
     * @param data integer array of allele data
     * @param hasMissing <code>true</code> if denotes a missing allele</code>
     * and <code>false</code> if there is no missing allele code.
     */ 
    private void setGenotypeData(int refAllele, int[] data, int[] strata, 
            boolean hasMissing) {
        resetCounts();
        for  (int j=1; j<data.length; j+=2) {
            int jm1 = j - 1;
            int a1 = data[jm1];
            int a2 = data[j];
            boolean allele1 = (data[jm1] == refAllele) ^ !this.refIsMinor;
            boolean allele2 = (data[j] == refAllele) ^ !this.refIsMinor;
            int stratum = strata[j];
            if (hasMissing && (a1==0 || a2==0))  {
                missingGenotypes[stratum][missingGCount[stratum]++] = jm1;
                if (a1==0) {
                    missingAlleles[stratum][missingACount[stratum]++] = jm1;
                }
                else if (allele1) {
                    minorAlleles[stratum][minorCount[stratum]++] = jm1;                    
                }
                if (a2==0) {
                    missingAlleles[stratum][missingACount[stratum]++] = j;
                } else if (allele2) {
                    minorAlleles[stratum][minorCount[stratum]++] = j;
                }
            }
            else if (allele1 && allele2) {
                minorAlleles[stratum][minorCount[stratum]++] = jm1;
                minorAlleles[stratum][minorCount[stratum]++] = j;
                rec[stratum][rCount[stratum]++] = jm1;
                dom[stratum][dCount[stratum]++] = jm1;
            }
            else if (allele1) {
                minorAlleles[stratum][minorCount[stratum]++] = jm1;
                dom[stratum][dCount[stratum]++] = jm1;
            }
            else if (allele2) {
                minorAlleles[stratum][minorCount[stratum]++] = j;
                dom[stratum][dCount[stratum]++] = jm1;
            }
        }      
    }
    
    /*
     * Allocates the following two-dimensional arrays so that
     * <code>this.minorAlleles[j]</code>, 
     * <code>this.missingAlleles[j]</code>, <code>this.rec[j]</code>, 
     * <code>this.dom[j]</code>, and <code>this.missingGenotypes[j]</code> 
     * have size <code>this.minorCount[j]</code>, 
     * <code>this.rCount[j]</code>, <code>this.dCount[j]</code>, 
     * <code>this.missingACount[j]</code>, and <code>this.missingGCount[j]</code> 
     * for <code>0 &le; j < strataCounts.length</code>.
     *
     * @throws NegativeArraySizeException if
     * <code>this.minorCount[j] < 0</code>, 
     * <code>this.rCount[j] < 0</code>, <code>this.dCount[j] < 0</code>, 
     * <code>this.missingACount[j] < 0</code>, and 
     * <code>this.missingGCount[j] < 0</code> 
     * for <code>0 &le; j < nStrata</code>.
     */
    private void allocateArrays(int nStrata) {
        this.minorAlleles = new int[nStrata][];
        this.dom = new int[nStrata][];
        this.rec = new int[nStrata][];
        this.missingAlleles = new int[nStrata][];
        this.missingGenotypes = new int[nStrata][];
        for (int j=0; j<nStrata; ++j) {
            this.minorAlleles[j] = new int[minorCount[j]];
            this.rec[j] = new int[rCount[j]];
            this.dom[j] = new int[dCount[j]];
            this.missingAlleles[j] = new int[missingACount[j]];
            this.missingGenotypes[j] = new int[missingGCount[j]]; 
        }
    }
    
    /*
     * Allocates the private arrays <code>this.minorCount</code>, 
     * <code>this.missingACount</code>, <code>this.rCount</code>, 
     * <code>this.dCount</code>, and <code>this.missingGCount</code> 
     * to have the specified size.
     *
     * @throws NegativeArraySizeException if <code>size < 0</code>. 
     */
    private void allocateCountArrays(int size) {
        minorCount = new int[size];
        rCount = new int[size];
        dCount = new int[size];
        missingACount = new int[size];
        missingGCount = new int[size];        
    }
    
    /*
     * Resets the elements of the private arrays <code>minorCount</code>, 
     * <code>rCount</code>, <code>dCount</code>, 
     * <code>missingACount</code>, and <code>missingGCount</code> 
     * to 0.
     */
    private void resetCounts() {
        for (int j=0; j<strataCounts.length; ++j) {
            minorCount[j] = 0;
            rCount[j] = 0;
            dCount[j] = 0;
            missingACount[j] = 0;
            missingGCount[j] = 0;
        }
    }
    
    /**
     * Returns a string representation of <code>this</code>.   The 
     * exact details of the representation are unspecified and 
     * subject to change.
     *
     * @return a string representation of <code>this</code>.   The 
     * exact details of the representation are unspecified and 
     * subject to change.
     */
     public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append(" refAllele=" + refAllele());
        sb.append(" refIsMinor=" + refIsMinor);
        sb.append(" strataCounts=" + Arrays.toString(strataCounts()));
        sb.append(newLine);
        sb.append("strata=" + Arrays.toString(strata));
        sb.append(newLine);
        sb.append("minorAlleles = " + Arrays.toString(minorAlleles));
        sb.append(newLine);
        sb.append("dom = " + Arrays.toString(dom));
        sb.append(newLine);
        sb.append("rec = " + Arrays.toString(rec));
        sb.append(newLine);
        sb.append("missing alleles = " + Arrays.toString(missingAlleles));
        sb.append(newLine);
        sb.append("missing genotypes = " + Arrays.toString(missingGenotypes));
        return sb.toString();
    }
     
    /**
     * Returns the reference allele. 
     * 
     * @return the reference allele.
     */    
    public int refAllele() {
         return refAllele;
    }
     
    /**
     * Returns the number of alleles in the sample.
     *
     * @return the number of alleles in the sample.
     */
    public int alleleCount() {
        return alleleCount;
    }
    
    
    /**
     * Returns an array whose k-th element is the the number of alleles 
     * in the k-th strata of the sample (including missing alleles).   
     * 
     * @return an array whose k-th element is the the number of alleles 
     * in the k-th strata of the sample (including missing alleles). 
     */
    public int[] strataCounts() {
        return strataCounts;
    }

    /**
     * Returns <code>true</code> if the reference allele is the minor allele
     * and <code>false</code> otherwise.
     *
     * @return <code>true</code> if the reference allele is the minor allele
     * and <code>false</code> otherwise.
     */
    public boolean refIsMinor() {
        return refIsMinor;
    }

    /**
     * Returns a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the minor alleles in
     * the k-th stratum. 
     *
     * @return a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the minor alleles in
     * the k-th stratum. 
     */
    public int[][] minorAlleles() {
        return minorAlleles;
    }

    /**
     * Returns a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the dominant genotypes in
     * the k-th stratum. 
     * Only the index of the first allele in the dominant genotype is given.
     * The dominant genotype is defined as a genotype that carries one
     * or two copies of the minor allele.  Returns an <code>int[][]</code>
     * array of length 0 if the number of alleles is odd.
     *
     * @return a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the dominant genotypes in
     * the k-th stratum. 
     */
    public int[][] dom() {
        return dom;
    }

    /**
     * Returns a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the recessive genotypes in
     * the k-th stratum.   Only the index of the first allele in the 
     * recessive genotype is given.
     * The recessive genotype is defined as the homozygous minor allele 
     * genotype.   Returns an array of length 0 if the number of alleles 
     * is odd.
     *
     * @return a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the recessive genotypes in
     * the k-th stratum. 
     */
    public int[][] rec() {
        return rec;
    }

    /**
     * Returns a two-dimensional <code>int</code> array 
     * whose k-th row is the indices of the missing alleles in
     * the k-th stratum. 
     *
     * @return a two-dimensional <code>int</code> array 
     * whose k-th row is the indices of the missing alleles in
     * the k-th stratum. 
     */
    public int[][] missingAlleles() {
        return missingAlleles;
    }

    /**
     * Returns a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the missing genotypes in
     * the k-th stratum.  Only the index of the 
     * first allele (smallest index) in each missing genotype 
     * is given.  Genotypes are considered missing if either one or 
     * both alleles are missing.  Returns an <code>int[][]</code> array 
     * of length 0 if the number of alleles is odd.
     *
     * @return a two-dimensional <code>int[][]</code> array 
     * whose k-th row is the indices of the missing genotypes in
     * the k-th stratum.
     */
    public int[][] missingGenotypes() {
        return missingGenotypes;
    }
}