/*
 * Diallelic.java
 *
 * Created on 28 May 2006, 17:29
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

/**
 * Class <code>Diallelic</code> represents sample data
 * for a diallelic marker.
 */
public interface Diallelic {
    
    /**
     * Returns the marker identifier.
     *
     * @return the marker identifier.
     */
    public String id(); 
    
    /**
     * Returns the reference allele.  For 
     * <code> 0 &le; k < this.alleleSize()</code>,
     * <code>this.data(k)</code> returns <code>true</code> if the 
     * <code>k</code>-th allele is the reference allele and false otherwise.
     * 
     * @return the reference allele.
     */
    public String refAllele();  

    /**
     * Returns the number of alleles in the sample (including missing alleles).   
     * 
     * @return the number of alleles in the sample (including missing alleles).
     */
    public int alleleCount();
    
    /**
     * Returns an array whose k-th element is the the number of alleles 
     * in the k-th strata of the sample (including missing alleles).   
     * 
     * @return an array whose k-th element is the the number of alleles 
     * in the k-th strata of the sample (including missing alleles). 
     */
    public int[] strataCounts();
    
    /**
     * Returns <code>true</code> if the <code>k</code>-th allele 
     * is the reference allele and <code>false</code> otherwise.
     * 
     * @return <code>true</code> if the <code>k</code>-th allele 
     * is the reference allele and <code>false</code> otherwise.
     * @throws IndexOutOfBoundsException if 
     * <code>k < 0 || k &ge; this.alleleCount()</code>.
     */
    public boolean data(int k);
    
    /**
     * Returns an array whose k-th element is the the number of alleles 
     * in the k-th strata that are minor alleles and whose indices
     * correspond to <code>true</code> entries in the specified boolean array.
     * If <code>counts.length == strataCounts().length</code> then
     * the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.
     * 
     * @param trait a boolean array of length <code>this.alleleCount()</code>.
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned
     * @return an array whose k-th entry is the the number of alleles 
     * in the k-th strata that are minor alleles and whose indices
     * correspond to <code>true</code> entries in the specified boolean array.
     *
     * @throws IllegalArgumentException if 
     * <code>trait.length != this.alleleCount()</code>.
     * @throws NullPointerException if 
     * <code>trait==null || counts==null</code>.
     */
    public int[] minorAlleleCount(boolean[] trait, int[] counts);
    
    /**
     * Returns an array whose k-th element is the number of genotypes in
     * the k-th strata that are recessive and whose indices correspond to 
     * <code>true</code> entries of the specified boolean array.  All
     * entries of the returned array are 0 if there are an odd number 
     * of alleles.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.  The recessive genotype is 
     * defined as the homozygous minor allele genotype.   The behaviour 
     * of this method is undefined if there exists a genotype in which 
     * the trait is not the same for both alleles in the genotype.
     * 
     * @param trait a boolean array of length 
     * <code>this.alleleCount()</code> that specifies a boolean value
     * for each allele.  The boolean trait must be the same for both
     * alleles in each genotype.
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned
     * @return an array whose k-th entry is the number of genotypes in
     * the k-th strata that are recessive and whose indices correspond to 
     * <code>true</code> entries of the specified boolean array.  All
     * entries of the returned array are 0 if there are an odd number 
     * of alleles.
     * @throws IllegalArgumentException if
     * <code>trait.length != this.alleleCount()</code>.
     * @throws NullPointerException if 
     * <code>trait==null || counts==null</code>.
     */
    public int[] recCount(boolean[] trait, int[] counts);
    
    /**
     * Returns an array whose k-th element is the number of genotypes in
     * the k-th strata that are dominant and whose indices correspond to 
     * <code>true</code> entries of the specified boolean array.  All
     * entries of the returned array are 0 if there are an odd number 
     * of alleles.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.  A dominant genotype is 
     * defined as the a genotype that carries one or two copies of the
     * minor allele.  The behaviour 
     * of this method is undefined if there exists a genotype in which 
     * the trait is not the same for both alleles in the genotype.
     * 
     * @param trait a boolean array of length 
     * <code>this.alleleCount()</code> that specifies a boolean value
     * for each allele.  The boolean trait must be the same for both
     * alleles in each genotype.
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th entry is the number of genotypes in
     * the k-th strata that are dominant and whose indices correspond to 
     * <code>true</code> entries of the specified boolean array.  All
     * entries of the returned array are 0 if there are an odd number 
     * of alleles.
     * @throws IllegalArgumentException if
     * <code>trait.length != this.alleleCount()</code>.
     * @throws NullPointerException if 
     * <code>trait==null || counts==null</code>.
     */
    public int[] domCount(boolean[] trait, int[] counts);

    /**
     * Returns an array whose k-th element is the number of alleles 
     * in the k-th strata that are missing and whose indices correspond to 
     * <code>true</code> entries of the specified boolean array. 
     * If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.
     * 
     * @param trait a boolean array of length 
     * <code>this.strataCounts()</code> that specifies a boolean value 
     * for each allele.
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th entry is the number of alleles 
     * that are missing and whose indices correspond to 
     * <code>true</code> entries of the specified boolean array. 
     *
     * @throws IllegalArgumentException if
     * <code>trait.length != this.alleleCount</code>.
     * @throws NullPointerException if 
     * <code>trait==null || counts==null</code>.
     */    
    public int[] missingAlleleCount(boolean[] trait, int[] counts);

    /**
     * Returns an array whose k-th element is the number of genotypes
     * in the k-th strata that are missing and whose indices correspond to the
     * <code>true</code> entries of the specified boolean array.
     * All entries of the returned array are 0 if there are an odd number 
     * of alleles.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated. 
     * A genotype is missing if either or both alleles are missing.
     * The behaviour of this method is undefined if there exists a genotype 
     * in which the trait is not the same for both alleles in the genotype.
     * 
     * @param trait a boolean array of length <code>this.alleleCount()</code> 
     * that specifies a boolean value for each allele.  The boolean value must 
     * be the same for both alleles in a genotype.
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th entry is the number of genotypes
     * in the k-th strata that are missing and whose indices correspond to the
     * <code>true</code> entries of the specified boolean array.
     * All entries of the returned array are 0 if there are an odd number 
     * of alleles. 
     *
     * @throws IllegalArgumentException if
     * <code>trait.length != this.alleleCount()</code>.
     * @throws NullPointerException if 
     * <code>trait==null || counts==null</code>.
     */        
    public int[] missingGenotypeCount(boolean[] trait, int[] counts);
    
    /**
     * Returns an array whose k-th element the number minor alleles 
     * in the k-th strata.  If 
     * <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.
     *
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return the number of non-missing minor alleles in the sample.
     * @throws NullPointerException if <code>counts==null</code>.
     */
    public int[] minorAlleleCount(int[] counts);
    
    /**
     * Returns an array whose k-th element is the number of 
     * dominant genotypes in the k-th strata.     All
     * entries of the returned array are 0 if there are an odd number 
     * of alleles.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.  A dominant genotype is 
     * defined as the a genotype that carries one or two copies of the
     * minor allele.
     * 
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th element is the number of 
     * dominant genotypes in the k-th strata.     All elements of
     * the returned array are 0 if there are an odd number of alleles. 
     * @throws NullPointerException if <code>counts==null</code>.
     */
    public int[] domCount(int[] counts);
    
    /**
     * Returns an array whose k-th element is the number of 
     * recessive genotypes in the k-th strata.     All
     * entries of the returned array are 0 if there are an odd number 
     * of alleles.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.  The recessive genotype is 
     * defined as the homozygous minor allele genotype.
     * 
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th element is the number of 
     * recessive genotypes in the k-th strata.     All elements of
     * the returned array are 0 if there are an odd number of alleles. 
     * @throws NullPointerException if <code>counts==null</code>.
     */
    public int[] recCount(int[] counts);
    
    /**
     * Returns an array whose k-th element is the number of missing 
     * alleles in the k-th strata. 
     * If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated.
     * 
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th entry is the number of missing 
     * alleles in the k-th strata.
     *
     * @throws NullPointerException if <code>counts==null</code>.
     */   
    public int[] missingAlleleCount(int[] counts); 
    
    /**
     * Returns an array whose k-th element is the number of missing 
     * genotypes in the k-th strata.
     * All entries of the returned array are 0 if there are an odd number 
     * of alleles.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array is allocated. 
     * A genotype is missing if either or both alleles are missing.
     * 
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned.
     * @return an array whose k-th entry is the number of missing
     * genotype in the k-th strata.  All entries of the returned array 
     * are 0 if there are an odd number of alleles. 
     *
     * @throws NullPointerException if <code>counts==null</code>.
     */     
    public int[] missingGenotypeCount(int counts[]);
    
    /**
     * Returns the diallelic marker obtained from <code>this</code> by 
     * deleting alleles with indices corresponding to <code>true</code> 
     * entries in the specified boolean array. The specified boolean 
     * array is not changed by this method.
     * 
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     * @return the diallelic marker obtained from <code>this</code> by 
     * deleting alleles with indices corresponding to <code>true</code> 
     * entries in the specified boolean array.
     * @throws IllegalArgumentException if  
     * <code>isExcluded.length != this.alleleCount()</code>.
     */
    public Diallelic subset(boolean[] isExcluded);
    
    /**
     * Returns a string representation of <code>this</code>.  The exact details 
     * of the representation are unspecified and subject to change.
     *
     * @return a string representation of <code>this</code>.
     */
    public String toString();
}
