/*
 * FlexTest.java
 *
 * Created on 20 May 2006, 14:17
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;

/**
 * Class <code>FlexTest</code> is an immutable class used to compute 
 * association test statistics for a diallelic marker and a binary trait.
 */
public final class FlexTest implements Test {

    private static final char[] testCodes = new char[] {'a', 'd', 'r', 'o', 't'};
    private final int testCount;
    private final int[] caseAlleleCount;
    private final int[] alleleCount;
    private final boolean[] tests = new boolean[5];
    private final boolean useGenotypic;
    
    private final int[] tempCounts;
    private final int[] n;
    private final int[] nCases;
    private final int[] nAlleles;
    private final int[] nCaseAlleles;
    private final int[] nRec;
    private final int[] nCaseRec;
    private final int[] nDom;
    private final int[] nCaseDom;
    private final int[] nHet;
    private final int[] nCaseHet;
    
    /**
     * Creates a new instance of FlexTest.  
     * 
     * @param code a string determing the tests to be performed.
     * @param nCaseAlleles the number of case alleles in each strata.
     * @param nControlAlleles the number of control alleles in each strata.
     *
     * @throws NullPointerException if 
     * <code>code==null || nCaseAlleles==null || nControlAlleles==null</code>.
     * @throws IllegalArgumentException if 
     * <code>nCaseAlleles.length != nControlAlleles.length</code>, if 
     * any entry of the specified arrays is negative or if 
     * <code>FlexTest.isValidTest(code)==false</code> .
     */
    public FlexTest(String code, int[] nCaseAlleles, int[] nControlAlleles) {
        checkAlleleCounts(nCaseAlleles, nControlAlleles);
        int nStrata = nCaseAlleles.length;
        this.caseAlleleCount = (int[]) nCaseAlleles.clone();
        this.alleleCount = new int[nStrata];
        for (int j=0; j<alleleCount.length; ++j) {
            this.alleleCount[j] = nCaseAlleles[j] + nControlAlleles[j];
        }
        this.tempCounts = new int[nStrata];
        this.n = new int[nStrata];
        this.nCases = new int[nStrata];
        this.nAlleles = new int[nStrata];
        this.nCaseAlleles = new int[nStrata];
        this.nRec = new int[nStrata];
        this.nDom = new int[nStrata];
        this.nHet = new int[nStrata];
        this.nCaseRec = new int[nStrata];
        this.nCaseDom = new int[nStrata];
        this.nCaseHet = new int[nStrata];
        this.testCount = setTestsArray(code);
        this.useGenotypic = (tests[1] || tests[2] || tests[3] || tests[4]);
        if  (useGenotypic) {
            checkOdd(caseAlleleCount, alleleCount);
        }
    }
    
    /*
     * Checks that the specified arrays have the same length and have
     * no negative entries.
     *
     * @throws NullPointerException if 
     * <code>nCaseAlleles==null || nControlAlleles==null</code>.
     * @throws IllegalArgumentException if any entry of the 
     * specified arrays is negative.
     */
    private void checkAlleleCounts(int[] nCaseAlleles, int[] nControlAlleles) {
        if (nCaseAlleles.length != nControlAlleles.length) {
            String s = "nCaseAlleles.length" + nCaseAlleles.length + 
                    " != nControlAlleles.length=" + nControlAlleles.length;
            throw new IllegalArgumentException(s);
        }
        for (int j=0; j<nCaseAlleles.length; ++j) {
            if (nCaseAlleles[j] < 0) {
                String s = "negative array entry: ";
                throw new IllegalArgumentException(s + Arrays.toString(nCaseAlleles));
            }
            if (nControlAlleles[j] < 0) {
                String s = "negative array entry: ";
                throw new IllegalArgumentException(s + Arrays.toString(nCaseAlleles));                
            }
        }         
    }
    
    /**
     * Returns <code>true</code> if the specified code
     * contains at least one of the following characters
     * 'a', 'r', 'd', 'o', or 't' (case insensitive) and 
     * <code>false</code> otherwise.  The characters 
     * stand for allelic, recessive, dominant, and 
     * overdominant, and trend
     *
     * @return <code>true</code> if the specified code
     * contains at least one of the following characters:
     * 'a', 'r', 'd', or 'o', 't' (case insensitive) and 
     * <code>false</code> otherwise.
     */
    public static boolean isValidTest(String code) {
        code = code.toLowerCase();
        for (char ch : testCodes) {
            if (code.indexOf(ch) >= 0) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Sets an entry of the private boolean array <code>tests</code> to 
     * <code>true</code> if the corresponding association test
     * is to be performed and returns the number of tests
     * that will be performed.
     *
     * @param code a string determining the tests to be performed.
     * @return the number of tests that will be performed.
     *
     * @throws IllegalArgumentException  if 
     * <code>FlexTest.isValidTest(code)==false</code>.
     */
    private int setTestsArray(String code) {
        int cnt = 0;
        String testCode = code.toLowerCase();
        if (!isValidTest(testCode)) {
            String s = "Unrecognized test code (" + code;
            s += ").  Missing one or more of the following characters: a d r o t";
            throw new IllegalArgumentException(s);            
        }
        if (testCode.indexOf('a') >= 0) {
            ++cnt;
            tests[0] = true;
        }
        if (testCode.indexOf('r') >= 0) {
            ++cnt;
            tests[1] = true;
        }
        if (testCode.indexOf('d') >= 0) {
            ++cnt;
            tests[2] = true;    
        }
        if (testCode.indexOf('o') >= 0) {
            ++cnt;
            tests[3] = true;
        }
        if (testCode.indexOf('t') >= 0) {
            ++cnt;
            tests[4] = true;
        }
        assert cnt > 0;
        return cnt;
    }
    
    /**
     * Checks that each strata contains an even number of case alleles 
     * and control alleles.
     *
     * @param caseAlleleCount the number of cases alleles in each stratum.
     * @param alleleCount the number of alleles (case + control) in each stratum.
     *
     * @throws IllegalArgumentException if a stratum contains an odd number
     * of case alleles or an odd number of control alleles.
     */
    private void checkOdd(int[] caseAlleleCount, int alleleCount[]) {
        for (int j=0; j<alleleCount.length; ++j) {
            if ((caseAlleleCount[j] & 1) == 1) {
                String s = "Using genotypic test with odd number of alleles: ";
                throw new IllegalArgumentException(s + Arrays.toString(caseAlleleCount));
            }
            if ((alleleCount[j] & 1) == 1) {
                String s = "Using genotypic test with odd number of alleles: ";
                throw new IllegalArgumentException(s + Arrays.toString(alleleCount));
            }
        }
    }
    
    public double[] pValues(boolean[] trait, Diallelic marker) {
        double[] p = new double[testCount];
        int index = -1;
        if (tests[0]) {
            diff(alleleCount, marker.missingAlleleCount(tempCounts), n);
            diff(caseAlleleCount, marker.missingAlleleCount(trait, tempCounts), nCases);
            marker.minorAlleleCount(nAlleles);
            marker.minorAlleleCount(trait, nCaseAlleles);
            p[++index] = stratifiedChisqTest(nCaseAlleles, nCases, nAlleles, n);
        }
        if (useGenotypic) {
            diff2(alleleCount, marker.missingGenotypeCount(tempCounts), n);
            diff2(caseAlleleCount, marker.missingGenotypeCount(trait, tempCounts), nCases);
            marker.recCount(nRec);
            marker.recCount(trait, nCaseRec);
            marker.domCount(nDom);
            marker.domCount(trait, nCaseDom);
            diff(nDom, nRec, nHet);
            diff(nCaseDom, nCaseRec, nCaseHet);
            if (tests[1]) {
                p[++index] = stratifiedChisqTest(nCaseRec, nCases, nRec, n);
            }
            if (tests[2]) {
                p[++index] = stratifiedChisqTest(nCaseDom, nCases, nDom, n);
            }
            if (tests[3]) {
                p[++index] = stratifiedChisqTest(nCaseHet, nCases, nHet, n);
            }
            if (tests[4]) {
                p[++index] = stratifiedTrendTest(nCaseHet, nCaseRec, nHet, nRec, 
                        nCases, n);
            }
        }
        return p;
    }
    
    public double[] pValues(boolean[] trait, Diallelic marker, int[] caseAlleleCount, 
            int[] alleleCount) {
        double[] p = new double[testCount];
        int index = -1;
        if (tests[0]) {
            diff(alleleCount, marker.missingAlleleCount(tempCounts), n);
            diff(caseAlleleCount, marker.missingAlleleCount(trait, tempCounts), nCases);
            marker.minorAlleleCount(nAlleles);
            marker.minorAlleleCount(trait, nCaseAlleles);
            p[++index] = stratifiedChisqTest(nCaseAlleles, nCases, nAlleles, n);
        }
        if (useGenotypic) {
            diff2(alleleCount, marker.missingGenotypeCount(tempCounts), n);
            diff2(caseAlleleCount, marker.missingGenotypeCount(trait, tempCounts), nCases);
            marker.recCount(nRec);
            marker.recCount(trait, nCaseRec);
            marker.domCount(nDom);
            marker.domCount(trait, nCaseDom);
            diff(nDom, nRec, nHet);
            diff(nCaseDom, nCaseRec, nCaseHet);
            if (tests[1]) {
                p[++index] = stratifiedChisqTest(nCaseRec, nCases, nRec, n);
            }
            if (tests[2]) {
                p[++index] = stratifiedChisqTest(nCaseDom, nCases, nDom, n);
            }
            if (tests[3]) {
                p[++index] = stratifiedChisqTest(nCaseHet, nCases, nHet, n);
            }
            if (tests[4]) {
                p[++index] = stratifiedTrendTest(nCaseHet, nCaseRec, nHet, nRec, 
                        nCases, n);
            }
        }
        return p;
    }
    
//    // temp debug code
//    private static void printCounts(int[] nCaseAlleles, int[] nCases, int[] nAlleles, int[] n) {
//        System.out.println("FlexTest 272:");
//        System.out.println("nCaseAlleles: " + Arrays.toString(nCaseAlleles));
//        System.out.println("nCases:       " + Arrays.toString(nCases));
//        System.out.println("nAlleles:     " + Arrays.toString(nAlleles));
//        System.out.println("n:            " + Arrays.toString(n));
//        System.out.println();
//    }
//
//    // temp debug code
//    private static void printCounts(int[] nCaseHet, int[] nCaseRec, int[] nHet, 
//                int[] nRec, int[] nCases, int[] n) {
//        System.out.println("FlexTest 282:");
//        System.out.println("nCaseHet  " + Arrays.toString(nCaseHet));
//        System.out.println("nCaseRec: " + Arrays.toString(nCaseRec));
//        System.out.println("nHet:     " + Arrays.toString(nHet));
//        System.out.println("nRec:     " + Arrays.toString(nRec));
//        System.out.println("nCases:   "  + Arrays.toString(nCases));
//        System.out.println("n:        " + Arrays.toString(n));
//        System.out.println();
//    }
    


    
    
    /**
     * Stores <code>a[j] - b[j]</code> in <code>diff[j]</code>
     * for <code>0 &le; j < diff.length</code>.
     *
     * @throws NullPointerException if any of the specified arrays
     * are null.
     * @throws ArrayIndexOutOfBoundsException if 
     * <code>a.length < diff.length || b.length < diff.length</code>.
     */
    private void diff(int[] a, int[] b, int[] diff) {
        for (int j=0; j<diff.length; ++j) {
            diff[j] = a[j] - b[j];
        }
    }
    
    /**
     * Stores <code>a[j]/2 - b[j]</code> in <code>diff[j]</code>
     * for <code>0 &le; j < diff.length</code>.
     *
     * @throws NullPointerException if any of the specified arrays
     * are null.
     * @throws ArrayIndexOutOfBoundsException if 
     * <code>a.length < diff.length || b.length < diff.length</code>.
     */
    private void diff2(int[] a, int[] b, int[] diff) {
        for (int j=0; j<diff.length; ++j) {
            diff[j] = a[j]/2 - b[j];
        }
    }
    
    public String[] tests() {
        String[] descriptions = new String[testCount];
        int index=-1;
        if (tests[0]) {
            descriptions[++index] = "allelic_X2";
        }
        if (tests[1]) {
            descriptions[++index] = "rec_X2";
        }
        if (tests[2]) {
            descriptions[++index] = "dom_X2";
        }
        if (tests[3]) {
            descriptions[++index] = "overdom_X2";
        }
        if (tests[4]) {
            descriptions[++index] = "trend_X2";
        }
        return descriptions; 
    }

    /**
     * Returns the Cochran-Mantel-Haenszel chi-square test statistic 
     * (with continuity correction) for stratified 2 x 2 contingency tables.  
     * All specified arrays
     * must have the same length.  The indices for the specified arrays
     * correspond to distinct strata, and the length of the arrays determines 
     * the number of strata.  The contract for this method is undefined
     * if any element of the specified arrays is negative
     * 
     * @param nCaseFactors the number of nCases that are positive for 
     * the factor in each strata.
     * @param nCases the number of nCases in each strata.
     * @param nFactors the number of individuals (cases plus controls)
     * that are positive for the factor in each strata.
     * @param n the number of individuals (cases plus controls) in each strata.
     * @return the Cochran-Mantel-Haenszel chi-square test statistic 
     * (with continuity correction) for stratified 2 x 2 contingency tables.
     * @throws NullPinterException if any argument is <code>null</code>.
     * @throws IndexOutOfBoundsException if the specified arrays are
     * not the same length;
     */
    public static double stratifiedChisqTest(int[] nCaseFactors, int[] nCases, 
            int[] nFactors, int[] n) {
        checkArrayLengths(nCaseFactors, nCases, nFactors, n);
        double sqrtNum = 0.0;
        double den = 0.0;
        for (int j=0; j<n.length; ++j) {
            double nn = (double) n[j];
            double caseProp = nCases[j]/nn;
            double prop = caseProp * (nFactors[j]/nn);
            sqrtNum += (nCaseFactors[j] - (caseProp * nFactors[j]));
            den += (n[j] - nFactors[j]) * ((n[j] - nCases[j])/(nn - 1.0)) * prop;
        } 
        sqrtNum = Math.abs(sqrtNum);
        if (sqrtNum > 0.5) {        // continuity correction implemented in R
            sqrtNum -= 0.5;
        }
        if (den<=0) {
            return 0.0;
        }     
        return sqrtNum*sqrtNum/den;
    }
    
    /**
     * Returns the Mantel trend test statistic for stratified 2 x 3 
     * contingency tables with case-control status coded as 0/1 and 
     * allele dosage coded as 0/1/2.  All specified arrays
     * must have the same length.  The indices for the specified arrays
     * correspond to distinct strata, and the length of the arrays determines 
     * the number of strata.  The contract for this method is undefined
     * if any of the specified arrays has a negative element.
     * 
     * @ref Mantel N (1963) Chi-square tests with one degree of freedom: 
     * extensions of the Mantel-Haenszel procedure. J Am Stat Assoc 58:690-700.
     *
     * @param nRec the number of recessive genotypes in each stratum.
     * @param nHet the number of heterozygous genotypes in each stratum
     * @param nCaseRec the number of recessive genotypes in cases in each
     * stratum.
     * @param nCaseHet the number of heterozygous genotypes in cases
     * in each stratum.
     * @param nCases the number of cases in each stratum.
     * @param n the number of individuals (cases plus controls) in 
     * each stratum.
     *
     * @return the Mantel trend test statistic for stratified 2 x 3 
     * contingency tables with case-control status coded as 0/1 and 
     * allele dosage coded as 0/1/2.
     * @throws NullPinterException if any argument is <code>null</code>.
     * @throws IndexOutOfBoundsException if the specified arrays are
     * not the same length;
     */
    public static double stratifiedTrendTest(int[] nCaseHet, int[] nCaseRec,
            int[] nHet, int[] nRec, int[] nCases, int[] n) {
        checkArrayLengths(nCaseHet, nCaseRec, nHet, nRec, nCases, n);
        double den = 0.0;
        double sqrtNum = 0.0;
        for (int j=0; j<n.length; ++j) {
            double s = nHet[j] + 2.0*nRec[j];
            double ss = nHet[j] + 4.0*nRec[j];
            double t = (double) n[j];
            double propCases = nCases[j] / t;
            double propControls = (n[j] - nCases[j])/(t - 1.0);
            sqrtNum += nCaseHet[j] + 2.0*nCaseRec[j] - propCases*s;
            den += (propCases*propControls)*(ss - s*s/t);
        }
        if (den==0) {
            return 0.0;
        }           
        return sqrtNum*sqrtNum/den;
    }
    
    /*
     * Checks that all specified arrays have the same length.
     *
     * @throws NullPinterException if any argument is <code>null</code>.
     * @throws IndexOutOfBoundsException if the specified arrays are
     * not the same length;
     */
    private static void checkArrayLengths(int[] nCaseFactors, int[] nCases, 
            int[] nFactors, int[] n){
        int a = n.length;
        int b = nCases.length;
        int c = nFactors.length;
        int d = n.length;
        if (a != b || b != c || c != d) {
            String s = "Array parameters must have same length: ";
            s += a + " " + b + " " + c + " " + d;
            throw new IllegalArgumentException(s); 
        }
    }
    
    /*
     * Checks that all specified arrays have the same length.
     *
     * @throws NullPinterException if any argument is <code>null</code>.
     * @throws IndexOutOfBoundsException if the specified arrays are
     * not the same length;
     */    
    private static void checkArrayLengths(int[] nCaseHet, int[] nCaseRec,
            int[] nHet, int[] nRec, int[] nCases, int[] n) {
        int a = nRec.length;
        int b = nHet.length;
        int c = nCaseRec.length;
        int d = nCaseHet.length;
        int e = nCases.length;
        int f = n.length;
        if (a != b || b != c || c != d || d != e || e != f) {
            String s = "Array parameters must have same length: ";
            s += a + " " + b + " " + c + " " + d;
            throw new IllegalArgumentException(s); 
        }
    }
    
    /**
     * Returns the chi-square test statistic (with continuity correction) 
     * for a 2 x 2 contingency table.
     * 
     * @param nCaseFactors the number of nCases that are positive for 
     * some factor.
     * @param nCases the number of nCases
     * @param nFactors the number of nCases and controls that are positive for 
     * some factor.
     * @param n the number of nCases and controls.
     * @return the chi-square test statistic (with continuity correction) 
     * for a 2 x 2 contingency table.
     */
    public static double chisqTest(int nCaseFactors, int nCases, int nFactors, int n) {
        double den = (double) nCases * nFactors * (n - nCases) * (n - nFactors);
        if (den==0) {
            return 0.0;
        }
        int n11 = nCaseFactors;
        int n12 = nFactors - nCaseFactors;
        int n21 = nCases - nCaseFactors;
        int n22 = n - nFactors - n21;        
        
        double a = Math.abs(n11*n22 - n12*n21) - (0.5)*n;
        double num = n*a*a;

        return num/den;
    }
    
    
    /**
     * Returns the trend-test chi-square statistic for the specified
     * table with row-index and col-index weights.  Returns 0.0
     * if <code>table.length==0</code>.  The returned value is undefined
     * if any entry of the specified table is negative.
     *
     * @param a two dimensional contingency table.
     * @throws NullPointerException if <code>table==null</code>
     */
    public static double trendTest(int[][] table) {
        double sum = 0.0;
        double xSum = 0.0;
        double ySum = 0.0;
        double xySum = 0.0;
        double xxSum = 0.0;
        double yySum = 0.0;
        for (int i=0; i<table.length; ++i) {
            int[] row = table[i];
            for (int j=0; j<row.length; ++j) {
                double value = row[j];
                double iXvalue = i*value;
                double jXvalue = j*value;
                sum += value;
                xSum += iXvalue;
                ySum += jXvalue;
                xySum += j * iXvalue;
                xxSum += i * iXvalue;
                yySum += j * jXvalue;
            }
        }
        if (sum==0) {
            return 0.0;
        }
        double xMean = xSum/sum;
        double yMean = ySum/sum;
        double sqrtNum = xySum - xSum*yMean;
        double den = (xxSum - xSum*xMean)*(yySum - ySum*yMean);
        if (den==0) {
            return 0.0;
        }
        return sum*sqrtNum*sqrtNum/den;
    }
       
    public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append(Arrays.toString(tests));
        sb.append(" : ");
        sb.append(Arrays.toString(tests()));
        sb.append(newLine);
        sb.append("caseAlleleCount=");
        sb.append(Arrays.toString(caseAlleleCount));
        sb.append(newLine);
        sb.append("alleleCount=");
        sb.append(Arrays.toString(alleleCount));
        sb.append(newLine);
        return sb.toString();
    }     
}
