/*
 * InputIterator.java
 *
 * Created on 5 May 2006, 14:56
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;
import blbutil.*;


/**
 * Class <code>InputIterator</code> is used to iterate through 
 * single-marker tests determined by markers in a BEAGLE format
 * input file. Markers with <code>n > 2</code> nAlleles are transformed into
 * <code>n</code> separate diallelic markers defined by presence/absence
 * of each allele.
 */
public class InputIterator implements FileIterator<Diallelic> {
    
    private Diallelic next = null;
    private int[] strata = null;
    private MarkerIterator markerIt = null;
    private Queue<Diallelic> queue = new LinkedList<Diallelic>();

    /**
     * Creates a new instance of InputIterator.
     * 
     * @param beagleFile name of file in BEAGLE input format.
     * @param missingCode code for missing marker alleleles in BEAGLE file.
     * @param strata an array on nonnegative integers.  The k-th element 
     * of the strata array is the stratum for the k-th allele in the 
     * specified BEAGLE file.
     *
     * @throws NullPointerException if 
     * <code>beagleFile==null || strata==null</code>.
     */
    public InputIterator(String beagleFile, String missingCode, int[] strata) {
        if (strata==null) {
            throw new NullPointerException("strata==null");
        }
        this.markerIt = new MarkerIterator(beagleFile, missingCode);
        this.strata = (int[]) strata.clone();
        this.next = getNext();
    }
    
    public boolean hasNext() {
         return next!=null;
    }
        
    public Diallelic next() {
        if (next==null) {
            throw new NoSuchElementException();
        }
        Diallelic current = next;
        next = getNext();
        return current;
    }

    public void remove() {
        throw new UnsupportedOperationException();
    }
    
    public void close() {
        markerIt.close();
        next=null;
    }  
    
    public String toString() {
        String space = " ";
        StringBuilder sb = new StringBuilder();
        sb.append(space);
        sb.append("next=" + next);
        sb.append(space);
        sb.append("queue=" + queue);
        sb.append(space);
        sb.append("markerIt=" + markerIt);
        return sb.toString();
    } 
    
    /*
     * Returns the next diallelic marker.  Markers with <code>n > 2</code>
     * alleles, are split into <code>n</code> diallelic markers.
     *
     * @return the next diallelic marker.
     */
    private Diallelic getNext() {
        Diallelic diallelic = queue.poll();
        while (diallelic==null && markerIt.hasNext()) {
            MarkerLineInterface markerLine = markerIt.next();
            int start = 0;
            if (markerLine.hasMissingCode()) {
                ++start;
            }
            int nAlleleCodes = markerLine.nAlleleCodes();
            if ((nAlleleCodes - start) == 2) {
                --nAlleleCodes;   // test only one allele of diallelic marker;
            } 
            if ((nAlleleCodes - start) == 0) {  // all alleles are missing)
                queue.offer(new IntDiallelic(markerLine, 1, strata));
            }
            for (int k=start; k<nAlleleCodes; ++k) {
                queue.offer(new IntDiallelic(markerLine, k, strata));
            }
            diallelic = queue.poll();
        }
        return diallelic;
    }
}
