/*
 * IntDiallelic.java
 *
 * Created on 22 July 2006, 06:45
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;

/**
 * Class <code>IntDiallelic</code> is an immutable class 
 * representing sample data for a diallelic marker.
 */
public final class IntDiallelic implements Diallelic {
    /* 
     * This implementation uses sorted int arrays to store allele indices.
     */
    private final String id;
    private final String refAllele;
    private final int alleleCount;
    private final int[] strata;
    private final int[][] minorAlleles;
    private final int[][] dom;
    private final int[][] rec;
    private final int[][] missingAlleles;
    private final int[][] missingGenotypes;
    private final int[] strataCounts;
    private final boolean refIsMinor;
    
    
    /**
     * Creates a new instance of <code>IntDiallelic</code> representing
     * the specified allele of the specified genetic marker data.
     * 
     * @param markerLine genetic marker data.
     * @param refAllele the integer coding for the reference allele.
     * @param strata an <code>int</code> array whose k-th element gives
     * the population stratum of the k-th allele.
     *
     * @throws IllegalArgumentException if 
     * <code>strata.length != markerLine.data().length</code>
     * @throws IndexOutOfBoundsException if 
     * <code>refAllele < 0</code> or if element of the specified 
     * <code>strata</code> array is negative.
     * @throws NullPointerException if 
     * <code>markerLine==null || strata==null</code>.
     */    
    public IntDiallelic(MarkerLineInterface markerLine, int refAllele, 
            int[] strata) {
        int[] data = markerLine.data();
        this.id = markerLine.identifier();
        this.strata = (int[]) strata.clone();
        if (refAllele >= markerLine.nAlleleCodes()) {
            this.refAllele = "missing";
        }
        else {
            this.refAllele = markerLine.allele(refAllele);
        }
        AlleleIndices alleleIndices = 
                new AlleleIndices(markerLine, refAllele, strata);
        this.refIsMinor = alleleIndices.refIsMinor();
        this.alleleCount = alleleIndices.alleleCount();
        assert this.alleleCount == strata.length;
        this.minorAlleles = alleleIndices.minorAlleles();
        this.dom = alleleIndices.dom();
        this.rec = alleleIndices.rec();
        this.missingAlleles = alleleIndices.missingAlleles();
        this.missingGenotypes = alleleIndices.missingGenotypes();
        this.strataCounts = alleleIndices.strataCounts();
    }   
    
    public String id() {
        return id;
    }
    
    public String refAllele() {
        return refAllele;
    }
    
    public int alleleCount() {
        return alleleCount;
    }
    
    public int[] strataCounts() {
        return (int[]) strataCounts.clone();
    }

    public boolean data(int k) {
        for (int[] ia : minorAlleles) {
            if (Arrays.binarySearch(minorAlleles, k) >= 0) {
                return refIsMinor;
            }
        }
        return !refIsMinor;
    }
    
    public int[] minorAlleleCount(boolean[] trait, int[] counts) {
        return andCount(minorAlleles, trait, counts);
    }
    
    public int[] recCount(boolean[] trait, int[] counts) {
        return andCount(rec, trait, counts);
    }
        
    public int[] domCount(boolean[] trait, int[] counts) {
        return andCount(dom, trait, counts);
    }

    public int[] missingAlleleCount(boolean[] trait, int[] counts) {
        return andCount(missingAlleles, trait, counts);
    }
    
    public int[] missingGenotypeCount(boolean[] trait, int[] counts) {
        return andCount(missingGenotypes, trait, counts);
    }

    /*
     * Returns an array whose <code>j</code>-th element
     * is <code>indicies[j].length</code>  The returned
     * array is the specified <code>counts</code> array
     * if <code>indices.length == counts.length</code>.
     */
    private int[] lengths(int[][] indices, int[] counts) {
        if (counts.length != indices.length) {
            counts = new int[indices.length];
        }
        for (int j=0; j<counts.length; ++j) {
            counts[j] = indices[j].length;
        }
        return counts;        
    }
        
    public int[] minorAlleleCount(int[] counts) {
        return lengths(minorAlleles, counts);
    }
     
    public int[] domCount(int[] counts) {
        return lengths(dom, counts);
    }
    
    public int[] recCount(int[] counts) {
        return lengths(rec, counts);
    }
    
    public int[] missingAlleleCount(int[] counts) {
        return lengths(missingAlleles, counts);
    }
    
    public int[] missingGenotypeCount(int[] counts) {
        return lengths(missingGenotypes, counts);
    }
    
    public String toString() {
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append("id=" + id);
        sb.append(" refAllele=" + refAllele());
        sb.append(" refIsMinor=" + refIsMinor);
        sb.append(" strataCounts=" + Arrays.toString(strataCounts()));
        sb.append(newLine);
        sb.append("strata=" + Arrays.toString(strata));
        sb.append(newLine);
        sb.append("minorAlleles = " + raggedArrayToString(minorAlleles));
        sb.append(newLine);
        sb.append("dom = " + raggedArrayToString(dom));
        sb.append(newLine);
        sb.append("rec = " + raggedArrayToString(rec));
        sb.append(newLine);
        sb.append("missing alleles = " + raggedArrayToString(missingAlleles));
        sb.append(newLine);
        sb.append("missing genotypes = " + raggedArrayToString(missingGenotypes));
        return sb.toString();
    }
    
    /*
     * Returns a string representation of the specified ragged 
     * two-dimensional array.  The exact details of the 
     * representation are unspecified and subject to change.
     */
    private String raggedArrayToString(int[][] a) {
        StringBuilder sb = new StringBuilder(500);
        for (int[] ia : a) {
            sb.append(Arrays.toString(ia));
            sb.append("  ");
        }
        return sb.toString();
    }
    
    /*
     * Returns an array whose j-th entry is the number of elements in the 
     * j-th row of the specified two-dimensional indices 
     * array that are indices of <code>true</code> entries in the specified 
     * boolean array.  If <code>counts.length == strataCounts().length</code> 
     * then the returned array is the specified <code>counts</code> array; 
     * otherwise, a new array of length <code>strata.counts().length</code>
     * is allocated.
     *
     * @param indices a two-dimensional array whose elements in the k-th row
     * are indices of alleles in the k-th population stratum.
     * @param trait a boolean array of length <code>this.alleleCount()</code>.
     * @param counts the array which is returned if 
     * <code>counts.length == strataCounts().length</code>; otherwise
     * a new array is allocated and returned
     *
     * @return an array whose j-th entry is the number of elements in the 
     * j-th row of the specified two-dimensional indices 
     * array that are indices of <code>true</code> entries in the specified 
     * boolean array.
     *
     * @throws IndexOutOfBoundsException if any entry of the two-dimensional
     * <code>indices</code> array is greater than or equal to 
     * <code>trait.length</code>.
     *
     * @throws IllegalArgumentException if 
     * <code>trait.length != this.alleleCount</code>.
     */
    private int[] andCount(int[][] indices, boolean[] trait, int[] counts) {
        if (trait.length != this.alleleCount) {
            String s = "trait.length=" + trait.length + " != this.alleleCount="
                    + this.alleleCount;
            throw new IllegalArgumentException(s);
        }
        if (counts.length != indices.length) {
            counts = new int[indices.length];
        }
        for (int j=0; j<indices.length; ++j) {
            int count = 0;
            int[] ia = indices[j];
            for (int i : ia) {
                if (trait[i]) {
                    ++count;
                }
            }
            counts[j] = count;
        }
        return counts;
    }  
    
    public Diallelic subset(boolean[] isExcluded) {
        return new IntDiallelic(this, isExcluded);
    }
    
    /**
     * Constructs a diallelic marker from the specified diallelic marker by 
     * deleting alleles with indices corresponding to <code>true</code> 
     * entries of the specified boolean array. The specified boolean 
     * array is not changed by this constructor.
     * 
     * @param diallelic an <code>IntDiallelic</code> marker.
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     * @return the diallelic marker obtained from <code>this</code> by 
     * deleting alleles with indices corresponding to <code>true</code> 
     * entries in the specified boolean array.
     * @throws IllegalArgumentException if  
     * <code>isExcluded.length != this.alleleCount()</code>.
     */
    private IntDiallelic(IntDiallelic diallelic, boolean[] isExcluded) {
        if (isExcluded.length != diallelic.alleleCount) {
            String s = "isExcluded.length=" + isExcluded.length 
                    + " != diallelic.alleleCount()" + diallelic.alleleCount;
            throw new IllegalArgumentException(s);
        }
        int[] newIndices = newIndices(isExcluded);
        this.id = diallelic.id;
        this.refAllele = diallelic.refAllele;
        this.refIsMinor = diallelic.refIsMinor;
        this.strata = subArray(diallelic.strata, isExcluded);
        this.alleleCount = this.strata.length;
        this.strataCounts = new int[diallelic.strataCounts.length];
        for (int i : this.strata) {
            ++this.strataCounts[i];
        }
        this.minorAlleles = subset(diallelic.minorAlleles, newIndices);
        this.dom = subset(diallelic.dom, newIndices);
        this.rec = subset(diallelic.rec, newIndices);
        this.missingAlleles = subset(diallelic.missingAlleles, newIndices);
        this.missingGenotypes = subset(diallelic.missingGenotypes, newIndices); 
    }

    /*
     * Returns an array of length <code>isExcluded.length</code> 
     * whose j-th entry is the number of <code>true</code>
     * entries in the first <code>j-1</code> entries of the specified
     * <code>isExcluded</code> array if <code>isExcluded[j]==true</code>
     * and whose j-th entry is -1 if <code>isExcluded[j]==false</code>.
     *
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     */
    private int[] newIndices(boolean[] isExcluded) {
        int[] newIndices = new int[isExcluded.length];
        int cnt = 0;
        for (int j=0; j<isExcluded.length; ++j) {
            if (isExcluded[j]) {
                newIndices[j] = -1;
            }
            else {
                newIndices[j] = cnt++;
            }
        }
        return newIndices;
    }    
    
    /**
     * Returns an array obtained by deleting all elements of the specified
     * array whose index j satisfies <code>isExcluded[j]==false</code>.
     *
     * @param array an <code>int</code> array.
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     *
     * @return an array obtained by deleting all elements of the specified
     * array whose index j satisfies <code>isExcluded[j]==false</code>.
     * @throws NullPointerException if 
     * <code>array==null || isExcluded==null</code>.
     * @throws IllegalArgumentException if 
     * <code>array.length != isExcluded.length</code>.
     */
    public static int[] subArray(int[] array, boolean[] isExcluded) {
        if (array.length != isExcluded.length) {
            String s = "array length mismatch: " + array.length + "!=" + 
                    isExcluded.length;
            throw new IllegalArgumentException(s);
        }
        int[] subArray = new int[countFalse(isExcluded)];
        int cnt = 0;
        for (int j=0; j<isExcluded.length; ++j) {
            if (!isExcluded[j]) {
                subArray[cnt++] = array[j];
            }
        }
        assert cnt == subArray.length;
        return subArray;    
    }
    
    /**
     * Returns an array obtained by deleting all elements of the specified
     * array whose index j satisfies <code>isExcluded[j]==false</code>.
     *
     * @param array an <code>int</code> array.
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     *
     * @return an array obtained by deleting all elements of the specified
     * array whose index j satisfies <code>isExcluded[j]==false</code>.
     * @throws NullPointerException if 
     * <code>array==null || isExcluded==null</code>.
     * @throws IllegalArgumentException if 
     * <code>array.length != isExcluded.length</code>.
     */ 
    public static boolean[] subArray(boolean[] array, boolean[] isExcluded) {
        if (array.length != isExcluded.length) {
            String s = "array length mismatch: " + array.length + "!=" +  
                    isExcluded.length;
            throw new IllegalArgumentException(s);
        }
        boolean[] subArray = new boolean[countFalse(isExcluded)];
        int cnt = 0;
        for (int j=0; j<isExcluded.length; ++j) {
            if (!isExcluded[j]) {
                subArray[cnt++] = array[j];
            }
        }
        assert cnt == subArray.length;
        return subArray;
    }
    
    /**
     * Returns the number of <code>false</code> entries in the specified
     * boolean aray.
     *
     * @param ba a boolean array.
     * @return the number of <code>false</code> entries in the specified
     * boolean aray.
     *
     * @throws NullPointerException if <code>ba==null</code>.
     */
    public static int countFalse(boolean[] ba) {
        int cnt = 0 ;
        for (boolean b : ba) {
            if (!b) {
                ++cnt;
            }
        }
        return cnt;
    }
    /*
     * Returns an array consisting of elements <code>indexMap[x]</code> 
     * that satisfy <code>indexMap[x] &ge; 0</code> where <code>x</code> 
     * is an element of the specified indices array.   Each element of 
     * the specified indices array must be an index of the specified
     * index map array.
     *
     * @param indices an array of array indices.
     * @param indexMap an array mapping index <code>j</code> to 
     * <code>indexMap[j]</code>.
     * @return an array consisting of elements <code>indexMap[x]</code> 
     * that satisfy <code>indexMap[x] &ge; 0</code> where <code>x</code> 
     * is an element of the specified indices array.
     *
     * @throws IndexOutOfBoundsException if an lement of the specified
     * indices array is not an index of the specified index map array.
     */
    private static int[] subset(int[] indices, int[] indexMap) {
        int cnt = 0;
        for (int i : indices) {
            if (indexMap[i] >= 0) {
                ++cnt;
            }
        }
        int[] newIndices = new int[cnt];
        cnt = 0;
        for (int i : indices) {
            int newIndex = indexMap[i];
            if (newIndex >= 0) {
                newIndices[cnt++] = newIndex;
            }
        }
        return newIndices;
    }
    
    /*
     * Returns a two-dimensional array whose j-th rows is 
     * <code>IntDiallelic.subset(indices[j], indexMap)</code>.
     */
    private static int[][] subset(int[][] indices, int[] indexMap) {
        int[][] newIndices = new int[indices.length][];
        for (int j=0; j<indices.length; ++j) {
            newIndices[j] = subset(indices[j], indexMap);
        }
        return newIndices;
    }
  
    /**
     * Returns a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     * Approximately half of the cases and half of the control individuals
     * will be selected for exclusion.
     * Alleles of diploid individuals are either both included 
     * or both excluded.  The contract for this method is undefined if
     * <code>isAllelicData==false</code> and there is any individual
     * whose two entries in the specified <code>trait</code> array
     * are not the same.
     *
     * @param trait a boolean array whose <code>true</code> entries
     * correspond to affected alleles.
     * @param isAllelicData <code>false</code> if each consecutive pair of 
     * alleles is an individual's genotype (diploid data) and 
     * <code>true</code> if each allele is from a distinct individual 
     * (haploid data).
     * @param seed a seed for a random number generator
     *
     * @return an array whose <code>true</code> entries denote excluded 
     * haplotypes. 
     *
     * @throws IllegalArgumentException if <code>isAllelicData==false</code>
     * and <code>trait.length</code> is odd.
     */
    public static boolean[] excludedArray(boolean[] trait, 
            boolean isAllelicData, long seed) {
        Random random = new Random(seed);
        boolean[] isExcluded = new boolean[trait.length];
        int[] caseIndices = toIndexArray(trait);
        int[] controlIndices = toIndexArray(complement(trait));
        randomize(caseIndices, isAllelicData, random);
        randomize(controlIndices, isAllelicData, random);
        for (int j=0, n=caseIndices.length/2; j<n; ++j) {
            isExcluded[caseIndices[j]] = true;
        }
        for (int j=0, n=controlIndices.length/2; j<n; ++j) {
            isExcluded[controlIndices[j]] = true;
        }
        return isExcluded;
    }    
    
    /*
     * Returns an integer array whose entries give the indices
     * corresponding to <code>true</code> entries in the specified
     * boolean array.
     */
    private static int[] toIndexArray(boolean[] array) {
        int cnt = 0;
        for (boolean b : array) {
            if (b) {
                ++cnt;
            }
        }
        int[] indexArray = new int[cnt];
        cnt = 0;
        for (int j=0; j < array.length; ++j) {
            if (array[j]) {
                indexArray[cnt++] = j;
            }
        }
        return indexArray;
    }
    
    /*
     * Randomly shuffles the elements of the specified indices array.  If 
     * <code>isAllelicData == false</code>, pairs of elements
     * <code>indices[2*j], indices[2*j + 1]</code>
     * (rather than single elements) are shuffled.
     *
     * @param indices an array to be shuffled
     * @param isAllelicData <code>false</code> if consecutive pairs of elements 
     * (rather than single elements) are to be randomly reordered and 
     * <code>true</code> otherwise.
     * @param random a random number generator.
     *
     * @throws IllegalArgumentException if 
     * <code>(indices.length & 1)==1 && (isAllelicData==false)</code>.
     */    
    private static void randomize(int[] indices, boolean isAllelicData, 
            Random random) {
        if (isAllelicData) {
            for (int i=indices.length-1; i>0; --i) {
                int j = random.nextInt(i+1);
                int temp = indices[j];
                indices[j] = indices[i];
                indices[i] = temp;
            }               
        }
        else {
            if ((indices.length & 1)== 1) {
                String s = "indices.length is odd: " + indices.length;
                throw new IllegalArgumentException(s);
            }
            for (int i=indices.length-2; i>0; i-=2) {
                int j = 2*random.nextInt((i+2)/2);
                int temp = indices[j];
                indices[j] = indices[i];
                indices[i] = temp;
                temp = indices[j+1];
                indices[j+1] = indices[i+1];
                indices[i+1] = temp;
            }      
        }
    }    
    
    /*
     * Returns the complement array of the specified boolean array.
     * If <code>comp</code> is the returned array, then
     * <code>comp[j] != array[j]</code>.
     *
     * @param array a boolean array.
     * @return the complement array of the specified boolean array.
     *
     * @throws NullPointerException if <code>array=null</code>.
     */
    private static boolean[] complement(boolean[] array) {
        boolean[] reversed = new boolean[array.length];
        for (int j=0; j<array.length; ++j) {
            if (array[j]==false)
                reversed[j] = true;
        }
        return reversed;
    }
    
    /*
     * Checks that internal integer arrays are consistent, and 
     * throws an assertion error if any inconsistencies are detected.
     */
    private void checkConsistency() {
        for (int[] ia : dom) {
            for (int i : ia) {
                assert (Arrays.binarySearch(minorAlleles, i) >= 0)
                        || (Arrays.binarySearch(minorAlleles, i+1) >=0);
            }
        }
        for (int[] ia : rec) {
            for (int i : ia) {
                assert (Arrays.binarySearch(minorAlleles, i) >= 0)
                        && (Arrays.binarySearch(minorAlleles, i+1) >=0);
            }
        }
        for (int[] ia : missingAlleles) {
            for (int i : ia) {
                assert Arrays.binarySearch(minorAlleles, i) < 0;
            }
        }
        for (int[] ia : missingGenotypes) {
            for (int i : ia) {
                assert (Arrays.binarySearch(missingAlleles, i) >= 0)
                        || (Arrays.binarySearch(missingAlleles, i+1) >=0);            
            }
        }
        if ((alleleCount & 1)==0) {
            for (int[] ia : minorAlleles) {
                for (int j=0; j<ia.length; ++j) {
                    int index = ia[j];
                    if ((index & 1)==1) {
                        --index;
                        if (ia[j] == (ia[j-1] + 1)) {
                            assert (Arrays.binarySearch(rec, index) >= 0);
                        }
                    }
                    assert (Arrays.binarySearch(dom, index) >= 0);   
                }
            }
            for (int[] ia : missingAlleles) {
                for (int j=0; j<missingAlleles.length; ++j) {
                    int index = ia[j];
                    if ((index & 1)==1) {
                        --index;
                    }
                    assert (Arrays.binarySearch(missingGenotypes, index) >= 0);            
                }
            }
        }
        else {
            for (int j=0; j<strataCounts.length; ++j) {
                assert dom[j].length==0 && rec[j].length==0 && missingGenotypes[j].length==0;
            }
        }
    }
}