/*
 * MarkerIterator.java
 *
 * Created on 15 June 2006, 15:44
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;
import blbutil.*;

/**
 * An <code>blbutil.FileIterator</code> for iterating through 
 * the lines of a data input file which contain phenotypic or 
 * sequence data.  Lines are white-space delimited.  If
 * the line contains at least one non-white space character 
 * and if the first white-space delimited field is "M", the data in the 
 * line will be used to construct a <code>MarkerLineInterface</code>
 * which can be returned by <code>MarkerIterator</code>; otherwise, 
 * the line will be skipped.  
 */
public class MarkerIterator implements FileIterator<MarkerLineInterface> {
    
    private static final String code = "M";
    private String missingCode;
    private LineIterator lineIt = null;
    private MarkerLineInterface next = null;

    
    /**
     * Creates a new instance of MarkerIterator from the specified 
     * data file.
     * 
     * @param dataFile the filename of the input file containing 
     * white-space delimited allele data in BEAGLE format.  A line of allele 
     * data has the format "M markerIdentifier a1 a2 a3 a4 ..." 
     * where a1, a2, ... are allele identifiers.  If the alleles
     * represent genotypes then a1, a2 are the genotype for the 
     * first individual, a3, a4 are the genotypes for the second
     * individual and so on.
     *
     * @param missingCode the code representing missing alleles.  If
     * <code>missingCode==null</code> then all alleles are non-missing
     * and there is no code for missing alleles.
     *
     * @throws NullPointerException if <code>dataFile==null</code>.
     */
    public MarkerIterator(String dataFile, String missingCode) {
        if (dataFile==null) {
            throw new NullPointerException("dataFile==null");
        }
        this.lineIt = new LineIterator(dataFile);
        this.missingCode = missingCode;
        next = nextLine();  
    }
    
    public boolean hasNext() {
        return (next != null);
    }
    
    public MarkerLineInterface next() {
        if (!hasNext()) {
            throw new NoSuchElementException();
        }
        MarkerLineInterface current = next;
        next = nextLine();
        return current;
    }
    
    public void remove() {
        throw new UnsupportedOperationException();
    }
    
    public void close() {
        lineIt.close();
    }
    
    /*
     * Returns a MarkerLine instance representing the next line of the 
     * datafile with marker data or <code>null</code> if no such line 
     * exists.
     */
    private MarkerLineInterface nextLine() {
        MarkerLineInterface nextLine = null;                
        while (lineIt.hasNext() && nextLine==null) {
            String line = lineIt.next().trim();
            if (!line.equals("")) {
                String[] fields = StringUtil.getFields(line, 3);
                if (fields[0].equals(code)) {
                    if (fields.length==1) {
                        String s = "No marker identifier in line: " + line;
                        throw new IllegalStateException(s);
                    }
                    else if (fields.length==2) {
                        String s = "No marker data in line: " + line;
                        throw new IllegalArgumentException(s);
                    }
                    else {
                        String id = new String(fields[1]);
                        nextLine = new MarkerLine(id, fields[2], missingCode);
                    }
                }

            }
        }
        return nextLine;          
    }
}
