/*
 * MarkerLine.java
 *
 * Created on 22 June 2006, 11:32
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;
import blbutil.*;

/**
 * Class <code>MarkerLine</code> represents marker data coded as 
 * integers.  
 */
public final class MarkerLine implements MarkerLineInterface {
        
    private final String identifier;
    private final List<String> coding = new ArrayList<String>(3);
    private final int[] alleles;
    private final boolean hasMissingCoding; 
    
   /**
    * Constructs a new <code>MarkerLine</code> instance from the
    * specified identifier, data line and missingCode code.  The white-space
    * delimited fields in the data line are string identifiers for 
    * alleles.   White space is defined as any unicode characters less 
    * than or equal to '\u0020' (the space character).
    * 
    * If <code>missingCode==null</code>, alleles are coded as 0, 1, 2, ... 
    * in the order they appear in the data line.  If 
    * <code>missingCode != null</code> then alleles that are equal to the 
    * non-null <code>missingCode</code>  are coded as 0, 
    * and all other alleles are coded as 1, 2, 3, ... in the order they
    * appear in the data line.  
    * 
    * @param identifier the identifier for the marker.
    * @param data a string of white-space delimited alleles.
    * @param missingCode a string representing a missing allele.
    * @throws NullPointerException if 
    * <code>identifier==null || dataLine==null</code>.
    */
    public MarkerLine(String identifier, String dataLine, String missingCode) {
        if (identifier==null) {
            throw new NullPointerException("identifier==null");
        }
        this.identifier = identifier;
        if (missingCode != null) {
            this.coding.add(missingCode);
            this.hasMissingCoding = true;
        }
        else {
            this.hasMissingCoding = false;
        }
        this.alleles = new int[StringUtil.countFields(dataLine)];
        int alleleCount = 0;
        int n = dataLine.length();
        int a = StringUtil.nextNonWhiteSpace(dataLine, 0);
        int b = (a < n) ? StringUtil.nextWhiteSpace(dataLine, a) : a;
        while (a < b) {
            alleles[alleleCount++] = indexOf(dataLine, a, b);
            a = (b < n) ? StringUtil.nextNonWhiteSpace(dataLine, b) : n;
            b = (a < n) ? StringUtil.nextWhiteSpace(dataLine, a) : n;
        }
        assert alleleCount==alleles.length;
        return;
    }    
    
    public boolean hasMissingCode() {
        return hasMissingCoding;
    }
    
    /**
     * the index of the added substring in <code>this.coding</code>
     * (after adding the specified substring to <code>this.coding</code>).
     *
     * @param line a string containing an allele code.
     * @param start the starting index (in line) of the substring (inclusive).
     * @param end the ending index (in line) of the substring (exclusive).
     *
     * @return the index of the added substring in <code>this.coding</code>
     * (after adding the specified substring to <code>this.coding</code>).
     */
    private int addCoding(String line, int start, int end) {
        char[] ca = new char[end - start];
        line.getChars(start, end, ca, 0);
        int index = coding.size();
        coding.add(new String(ca)); 
        return index;
    }
    
    /*
     * Returns the index of the specified substring in 
     * <code>this.coding</code>.   If the substring is not 
     * present in <code>this.coding</code>, then the  
     * substring will be added to the end of 
     * <code>this.coding</code>.
     *
     * @param line a string containing an allele code.
     * @param start the starting index (in line) of the substring (inclusive).
     * @param end the ending index (in line) of the substring (exclusive).
     *
     * @return the index of the specified substring in the 
     * coding list.  If the substring is not present in 
     * <code>this.coding</code>, then the substring will be added to the 
     * end of <code>this.coding</code>.
     */
    private int indexOf(String line, int start, int end) {
        int length = end - start;
        assert length > 0 && end <= line.length();
        for (int index=0, n=coding.size(); index<n; ++index) {
            String s = coding.get(index);
            if (s.length() == length) {
                int k=0;
                while (k<length && s.charAt(k)==line.charAt(start + k)) {
                    ++k;
                }
                if (k==length) {
                    return index;
                }
            }
        }
        return addCoding(line, start, end);
    }
    
    public String identifier() {
        return identifier;
    }
    
    public int[] data() {
        int[] copy = new int[alleles.length];
        System.arraycopy(alleles, 0, copy, 0, copy.length);
        return copy;
    }
    
    public int nAlleleCodes() {
        return coding.size();
    }
    
    public int nAlleles() {
        return alleles.length;
    }
    
    public String allele(int j) {
        return coding.get(j);
    }
    
    public String toString() {
        String space = "\t";
        String newLine = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();
        sb.append(identifier);
        for (int a : alleles) {
            sb.append(space);
            sb.append(a);
        }
        sb.append(newLine);
        sb.append(identifier);
        for (int a : alleles) {
            sb.append(space);
            sb.append(coding.get(a));
        }
        return sb.toString();
    }
}