/*
 * NullDistribution.java
 *
 * Created on 26 July 2006, 18:45
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.io.*;
import java.util.Random;
import java.util.Arrays;
import blbutil.*;

// xxx need to document methods

/**
 * Class <code>NullDistribution</code> represents the 
 * distributions of a fixed number of order statistics obtained 
 * from testing diallelic markers for association with a fixed 
 * sequence of permutations of the trait status.  The permutations
 * of the trait status are performed within population strata.
 */
public class NullDistribution {
    
    private final double threshold;
    private final Random random;
    
    private final boolean[] trait;
    private final int[][] strataIndices;
    private final Test test;
    private final boolean isAllelicData;
    
    private final int maxSavedRank;
    private final double[][] maxNullStatistics;
    private final boolean[][] allPerms;
    
    private final boolean[] isExcluded;
    private final boolean[][] stage1Perms;
    private final int[][] nStage1CaseAlleles;
    private final int[] nStage1Alleles; 
            
    /**
     * Creates a new instance of NullDistribution for the specified 
     * parameters.
     * 
     * @param trait an array whose <code>j</code>-th element is 
     * <code>true</code> if the <code>j</code>-th haplotype is affected and 
     * is <code>false</code> otherwise.
     * @param strata an array of length <code>trait.length</code> whose j-th 
     * element is a nonnegative integer giving the population strata of the 
     * j-th allele.
     * @param nPerms number of permutations.  Must be a non-negative integer.
     * @param maxSavedRank number of largest order statistics whose
     * distributions are represented by <code>this</code>.  Must be a postive 
     * integer.
     * @param seed seed for random number generator.
     * @param isAllelicData <code>true</code> if the alleles are to be 
     * randomly reordered, and <code>false</code> if the genotypes are to 
     * be randomly reordered.  If <code>false</code> is specified, 
     * the pairs trait[2*j], trait[2*j + 1] are kept together during 
     * reordering.
     * @param tests an <code>Test</code> object for testing
     * for associations between the permuted trait and diallelic
     * marker data.
     * @throws NullPointerException if 
     * <code>trait==null || test==null || strata==null</code>.
     * @throws IllegalArgumentException if 
     * <code>(nPerms < 0) || (maxSavedRank < 1) || 
     * (trait.length != strata.length) || 
     * (!isAllelicData && ((trait.length&1)==0))</code>.
     */
    public NullDistribution(boolean[] trait, int[] strata, int nPerms, 
            int maxSavedRank, long seed, boolean isAllelicData, Test test, 
            double threshold) {
        checkArguments(trait, strata, nPerms, maxSavedRank, isAllelicData, test, 
                threshold);
        this.trait = (boolean[]) trait.clone();
        this.strataIndices = strataIndices(strata); 
        this.random = new Random(seed);
        this.test = test;
        this.isAllelicData = isAllelicData;
        this.maxSavedRank = maxSavedRank;
        this.maxNullStatistics = new double[nPerms][maxSavedRank];
        this.threshold = threshold;
        this.allPerms = allPerms(trait, isAllelicData, nPerms);
        if (threshold > 0.0) {
            isExcluded = IntDiallelic.excludedArray(trait, isAllelicData, seed);
            stage1Perms = stage1Perms(allPerms, isExcluded);
            nStage1Alleles = stage1AlleleCounts(strata, isExcluded);
            int[] stage1Strata = IntDiallelic.subArray(strata, isExcluded);
            nStage1CaseAlleles = stage1CaseAlleleCounts(stage1Perms, stage1Strata);
        }
        else {
            isExcluded = null;
            stage1Perms = null;
            nStage1Alleles = null;
            nStage1CaseAlleles = null;
        }
    }
     
    /*
     * Checks arguments to <code>NullDistribution</code> constructor.
     */
    private void checkArguments(boolean[] trait, int[] strata, int nPerms, 
            int maxSavedRank, boolean isAllelicData, Test test, 
            double threshold) {
        if (trait==null) {
            throw new NullPointerException("trait==null");
        }
        if (test==null) {
            throw new NullPointerException("test==null");
        }
        if (strata==null) {
            throw new NullPointerException("strata==null");
        }
        if (trait.length != strata.length) {
            String s = "trait.length!=strata.length: ";
            s += trait.length + " " + strata.length;
            throw new IllegalArgumentException(s);
        }
        if (!isAllelicData && (trait.length & 1)!=0) {
            String s = "isAllelicData==false, but trait.length=" + trait.length;
            throw new IllegalArgumentException(s);
        }
        if (nPerms < 0) {
            throw new IllegalArgumentException("nPerms < 0: " + nPerms);
        }
        if (maxSavedRank < 1) {
            throw new IllegalArgumentException("maxSavedRank < 1: " + maxSavedRank);
        }     
    }
    
    /*
     * Returns a two-dimensional array with <code>nPerms</code> rows and 
     * <code>trait.length</code> columns such that each row corresponds
     * to a permutation of the trait status within strata.
     *
     * @param trait an array whose <code>j</code>-th element is 
     * <code>true</code> if the <code>j</code>-th haplotype is affected and 
     * is <code>false</code> otherwise.
     * @param isAllelicData <code>true</code> if the alleles are to be 
     * randomly reordered, and <code>false</code> if the genotypes are to 
     * be randomly reordered.  If <code>false</code> is specified, 
     * the pairs trait[2*j], trait[2*j + 1] are kept together during 
     * reordering.
     * @param nPerms number of permutations.  Must be a non-negative integer.
     *
     * @return a two-dimensional array with <code>nPerms</code> rows and 
     * <code>trait.length</code> columns such that each row corresponds
     * to a permutation of the trait status within strata.
     */
    private boolean[][] allPerms(boolean[] trait, boolean isAllelicData, 
            int nPerms) {
        assert strataIndices != null;
        boolean[][] allPerms = new boolean[nPerms][];
        boolean[] permTrait = (boolean[]) this.trait.clone();
        for (int j=0; j<allPerms.length; ++j) {
            randomize(permTrait, isAllelicData);
            allPerms[j] = (boolean[]) permTrait.clone();
        }
        return allPerms;
    }

    /*
     * Returns the two-dimensional permutation array after removing the 
     * specified alleles.
     *
     * @param allPerms a two-dimensional array where each row is a permutation
     * of the trait status.
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     */
    private boolean[][] stage1Perms(boolean[][] allPerms, boolean[] isExcluded) {
        int length = IntDiallelic.countFalse(isExcluded);
        int nPerms = allPerms.length;
        boolean[][] stage1 = new boolean[nPerms][length];
        for (int j=0; j<nPerms; ++j) {
            int index = 0;
            for (int k=0; k<isExcluded.length; ++k) {
                if (!isExcluded[k]) {
                    stage1[j][index++] = allPerms[j][k];
                }
            }
            assert index==stage1[j].length;
        }
        return stage1;
    }
    
    /*
     * Returns an array with whose k-th element is the number of alleles
     * in the k-th strata which are not excluded.
     *
     * @param strata an array of length <code>trait.length</code> whose j-th 
     * element is a nonnegative integer denoting the population strata of the 
     * j-th trait allele.
     * @param isExcluded a boolean array whose <code>j</code>-th entry is 
     * <code>true</code> if the <code>j</code>-th allele is to be excluded.
     */
    private int[] stage1AlleleCounts(int[] strata, boolean[] isExcluded) {
        int nStrata = PermTester.max(strata) + 1;
        int[] stage1Strata = IntDiallelic.subArray(strata, isExcluded);
        int[] stage1Count = new int[nStrata];
        for (int i : stage1Strata) {
            ++stage1Count[i];
        }
        return stage1Count;
    }
    

    /**
     * Returns a two-dimensional array with <code>stage1Perms.length</code>
     * rows and <code>PermTester.max(stage1Strata) + 1</code> columns, such that
     * the entry in the j-th row and k-th column is the number of 
     * <code>true</code> (i.e. affected) entries in the j-th row of 
     * <code>stage1Perms</code> which are in the k-th stage1Strata.
     * 
     * @param stage1Perms a two-dimensional array with <code>nPerms</code> 
     * rows and <code>stage1Strata.length</code> columns such that each row 
     * corresponds to a permutation of the trait status of stage1 alleles 
     * within stage1Strata.
     * @param stage1Strata an array whose j-th element is a nonnegative integer 
     * giving the population strata of the j-th allele in the stage1 sample
     */
    private int[][] stage1CaseAlleleCounts(boolean[][] stage1Perms, 
            int[] stage1Strata) {
        int nStrata = PermTester.max(stage1Strata) + 1;
        int[][] caseAlleleCounts = new int[stage1Perms.length][nStrata];
        for (int j=0; j<stage1Perms.length; ++j) {
            for (int k=0; k<stage1Strata.length; ++k) {
                if (stage1Perms[j][k]) {
                    ++caseAlleleCounts[j][stage1Strata[k]];
                }
            }
        }
        return caseAlleleCounts;
    }      
            
    /*
     * Randomly reorders the alleles or genotypes in each population strata 
     * of the specified boolean array.  If 
     * <code>this.isAllelicData==true</code>, the alleles are
     * randomly reordered, and if <code>this.isAllelicData==false</code>,
     * the genotypes are randomly reordered so that the 
     * boolean pairs trait[2*j], trait[2*j + 1] are kept together during 
     * reordering.  
     *
     * Note: the private this.strataIndices field must be set prior to
     * calling this method.
     *
     * @param trait a boolean array.
     * @param isAllelicData <code>true</code> if the alleles are to be 
     * randomly reordered, and <code>false</code> if the genotypes are to 
     * be randomly reordered. 
     */   
    private void randomize(boolean[] trait, boolean isAllelicData) {
        if (!isAllelicData) {
            assert (trait.length & 1) == 0;
            for (int[] s : strataIndices) {
                for (int i=s.length-2; i>0; i-=2) {
                    int j = 2*random.nextInt((i/2) + 1);
                    int a = s[i];
                    int b = s[j];
                    assert trait[a]==trait[a+1];
                    boolean temp = trait[a];
                    trait[a] = trait[b];
                    trait[b] = temp;
                    temp = trait[a+1];
                    trait[a+1] = trait[b+1];
                    trait[b+1] = temp;

                }
            }      
        }
        else {
            for (int[] s : strataIndices) {
                for (int i=s.length-1; i>0; --i) {
                    int j = random.nextInt(i+1);
                    int a = s[i];
                    int b = s[j];
                    boolean temp = trait[a];
                    trait[a] = trait[b];
                    trait[b] = temp;
                }
            }
        }
    }

    /*
     * Returns a two-dimensional array whose j-th row gives is a sorted 
     * array of indices of trait alleles in the j-th population strata.
     *
     * @param strata an array of length <code>this.trait.length</code> whose j-th 
     * element is a nonnegative integer denoting the population strata of the 
     * j-th trait allele.
     */
    private int[][] strataIndices(int[] strata) {
        int[] counts = new int[PermTester.max(strata) + 1];
        for (int s : strata) {
            ++counts[s];
        }
        int[][] strataIndices = new int[counts.length][];
        for (int j=0; j<strataIndices.length; ++j) {
            strataIndices[j] = new int[counts[j]];
        }
        Arrays.fill(counts, 0);
        for (int j=0; j<strata.length; ++j) {
            int s = strata[j];
            strataIndices[s][counts[s]++] = j;
        }
        return strataIndices;
    }

    
    /**
     * Tests for association between <code>this.nPerms()</code> 
     * permuted trait status and the specified diallelic marker.
     * The <code>this.maxSavedRanks()</code> largest test statistics
     * for each permutation obtained from all invocations of the 
     * <code>this.permTest()</code> method can be retrieved using the
     * <code>this.orderStatistics()</code> method.
     * 
     * @param marker a diallelic marker.
     */
    public void permTest(final Diallelic marker) {
        Diallelic firstStage = null;
        if (threshold > 0.0) {
            firstStage = marker.subset(isExcluded);
        }
        for (int k=0; k<allPerms.length; ++k) {
            boolean[] permTrait = allPerms[k];
            double[] stats = maxNullStatistics[k];
            double max = 0.0;
            if (firstStage != null) {
                double[] pValues = test.pValues(stage1Perms[k], firstStage, 
                        nStage1CaseAlleles[k], nStage1Alleles);
                max = PermTester.max(pValues);
                if (max >= threshold) {
                    max = PermTester.max(test.pValues(permTrait, marker));
                }
            }
            else {
                max = PermTester.max(test.pValues(permTrait, marker));
            }
            if (max > stats[0]) {
                int index = Arrays.binarySearch(stats, max);
                if (index < 0)  {
                    index = -(index + 1);
                }
                int length = index - 1;
                System.arraycopy(stats, 1, stats, 0, length);
                stats[length] = max;
            }
        }
    }
    

    /**
     * Returns the number of permutations of the trait status that 
     * are tested against the marker data. 
     *
     * @return the number of permutations of the trait status that 
     * are tested against the marker data. 
     */
    public int nPerms() {
        return maxNullStatistics.length;
    }
    
    /**
     * Returns the number of order statistics 
     */
    public int maxSavedRank() {
        return maxSavedRank;
    }
    
    /**
     * Prints the null distribution of the order statistics to 
     * the specified file.  The output is tab delimited, and 
     * each line gives the top-ranked statistics observed for
     * each permutation of the trait status.  The permutations
     * determining each line are in same order as the permutations 
     * determining each element of the array returned by
     * the <code>orderStatistics()</code> method.
     * Each line contains the largest <code>this.maxSavedRank()</code>
     * chi-square statistics sorted in decreasing order.
     *
     * @param file an output file for the null distribution 
     * of order statistics.
     */
    public void printNullDistribution(File file) {
        String space = "\t";
        String newLine = System.getProperty("line.separator");
        String format = "%5.5e";
        PrintWriter nullOut = Utilities.createPrintWriter(file);
        for (double[] da : maxNullStatistics) {
            nullOut.print(String.format(format, da[da.length-1]));
            for (int k = da.length-2; k>=0; --k) {
                nullOut.print(space);
                nullOut.print(String.format(format, da[k]));
            }
            nullOut.println();
        }           
        nullOut.close(); 
    }
    
    /**
     * Returns a double array of length <code>this.nPerms()</code>
     * giving the specified order statistic for each permutation of the 
     * trait status.  The statistics are in order of decreasing 
     * significance (e.g. order 1 corresponds to the maximum 
     * chi-square statistic for each permutation of the trait status).
     * The <code>k</code>-th element of the returned 
     * array is the specified order statistic for the 
     * <code>k</code>-th permutation of the trait status.
     * 
     * @param order an integer denoting an order statistic satisfying
     * <code>1 &le; order &ge; this.maxSavedRank()</code>.
     * @return a double array of length <code>this.nPerms()</code>
     * giving the specified order statistic for each permutation of the 
     * trait status.
     * @throws IndexOutOfBoundsException if 
     * <code>order < 1 || order > this.maxSavedRank()</code>
     */
    public double[] orderStatistics(int order) {
        int index = maxSavedRank - order;
        double[] distribution = new double[maxNullStatistics.length];
        for (int k=0; k<maxNullStatistics.length; ++k) {
            distribution[k] = maxNullStatistics[k][index];
        }
        return distribution;
    }
    
    /**
     * Returns a copy of the unpermuted trait status.
     *
     * @return a copy of the unpermuted trait status.
     */
    public boolean[] trait() {
        return (boolean[]) trait.clone();
    }
    
    /**
     * Returns the <code>Test</code> object used for testing
     * for associations between the permuted trait and diallelic
     * marker data.
     *
     * @return the <code>Test</code> object used for testing
     * for associations between the permuted trait and diallelic
     * marker data.
     */
    public Test test() {
        return test;
    }
    
    /**
     * Returns the threshold for two-stage genotyping.  If the 
     * threshold is less than or equal to 0.0, one stage 
     * genotyping is assumed with all
     * samples genotyped in the first stage.
     *
     * @return the threshold for two-stage genotyping. 
     */
    public double threshold() {
        return threshold;
    }
}