/*
 * PValueWriter.java
 *
 * Created on 26 July 2006, 19:37
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.io.*;
import java.util.*;
import blbutil.*;

/**
 * A helper class for computing and writing permutation p-values.
 */
public final class PValueWriter {
    
    private final PrintWriter log;
    private final String header;
    private final double[] maxStats;
    private final double minMaxPermStat;
    private final double den;
    
    private int sigCount = 0;

    private static final double permLevel = 0.20;
    private static final String space = "\t";
    private static final String newLine = System.getProperty("line.separator");
    private static final String fDouble = "%11.4g";
    private static final String fStringDouble = "%11s";
    private static final String fMarker = "%-15s";
    private static final String fAllele = "%7s";
    
    /**
     * Creates a new instance of PValueWriter and writes marker test
     * statistics and permutation p-values to the specified 
     * <code>pValueFile</code>.
     * 
     * @param statisticsFile a file of statistics created with the 
     * <code>PValueWriter.writeStatistics</code> method.
     * @param pValueFile output file of statistics and permutation p-values.
     * @param log a <code>PrintWriter</code> for writing a summary of the 
     * permutation test.
     * @param nullDist null distribution of maximum test statistics 
     * (maximized over all markers) determined by the permuted trait status.
     * @throws NullPointerException if 
     * <code>statisticsFile==null || pValueFile || log==null 
     * || nullDist==null</code>.
     */
    public PValueWriter(File statisticsFile, String pValueFile, 
            PrintWriter log, NullDistribution nullDist) {
        this.log = log;
        this.maxStats = nullDist.orderStatistics(1);
        Arrays.sort(maxStats);
        boolean includePermPValue = (nullDist.nPerms() > 0);
        this.header = header(nullDist.test().tests(), includePermPValue);
        if (maxStats.length > 0) {
            this.minMaxPermStat =  maxStats[0];
            this.den = (double) maxStats.length + 1.0;     
        }
        else {
            this.minMaxPermStat = Double.NaN;
            this.den = Double.NaN;
        }
        PrintWriter out = Utilities.createPrintWriter(pValueFile);
        LineIterator it = new LineIterator(statisticsFile);
        out.println(header);
        while (it.hasNext()) {
            writePValues(it.next(), out);
        }
        it.close();
        out.close();
        
        if (maxStats.length > 0) {
            String alleles = (sigCount==1) ? " allele " : " alleles ";
            String s = newLine + sigCount + alleles + 
                    "with permutation p-value < " + permLevel;
            log.println(s);
        }
    }
    
    /*
     * Returns a header line for the p-value file reporting statistics
     * and permutation p-value for each marker tested.
     *
     * @return a header line for the p-value file reporting statistics
     * and permutation p-value for each marker tested.
     */
    private String header(String[] tests, boolean includePermPValue) {
        StringBuilder sb = new StringBuilder();
        sb.append(String.format(fMarker, "Marker"));
        sb.append(space);
        sb.append(String.format(fAllele, "Allele"));
        sb.append(space);
        for (String s : tests) {
            sb.append(String.format(fStringDouble, s));
            sb.append(space);
        }
        sb.append(String.format(fStringDouble, "max_X2"));
        if (includePermPValue) {
            sb.append(space);
            sb.append(String.format(fStringDouble, "perm_p_value"));
        }
        return sb.toString();
    }
        
    /*
     * Writes statistics and permutation p-values to the specified
     * print writer.  The statistics string must be created by the 
     * <code>writeStatistics</code> method.  The last field of
     * the specified statistics string (the maximum statistic) is used 
     * to compute the permutation p-value.
     *
     * @param statistics a line of test statistics written to a file with the 
     * <code>writeStatistics</code> method.
     * @param out a print writer for writing the test statistics and 
     * the permutation p-values.
     */
    private void writePValues(String statistics, PrintWriter out) {
        StringBuilder sb = new StringBuilder(120);
        String[] fields = StringUtil.getFields(statistics);
        sb.append(String.format(fMarker, fields[0]));
        sb.append(space);
        sb.append(String.format(fAllele, fields[1]));
        double p = -1.0;
        for (int j=2; j<fields.length; ++j) {
            sb.append(space);
            p = Double.parseDouble(fields[j]);
            sb.append(String.format(fDouble, p));
        }
        double max = p;
        if (maxStats.length > 0) {
            double permP = (max > minMaxPermStat) ? permutationPValue(max) : 1.0;
            sb.append(space);
            sb.append(String.format(fDouble, permP));
            if (permP < 0.20) {
                if (++sigCount==1) {
                    log.println("Alleles with permutation p-values less than 0.2:");
                    log.println();
                    log.println(header);
                }
                log.println(sb.toString());
            }
        }
        out.println(sb.toString());
    }
    
    /*
     * Returns the permutation p-value for the specified maximum statistic.
     * See Julian Besag and Peter Clifford (1991) Biometrika 78(2):301-4
     *
     * @param max a maximum test statistic for a marker
     */
    private double permutationPValue(double max) {
        int index = Arrays.binarySearch(maxStats, max);
        if (index < 0) {
            return ((maxStats.length + index + 2) / den);
        }
        else {
            while (--index >=0 && maxStats[index] >= max)
                ;
            return (maxStats.length - index) / den;
        } 
    }
    
    /**
     * Writes a white-space delimited line whose fields
     * (in ordera) are: a marker identifier, a reference allele for
     * that marker, the elements of an array of p-values, and 
     * the minimum p-value from the list of p-values.  
     * 
     * @param m a diallelic marker.
     * @param statistics an array of p-values.
     * @param out a <code>PrintWriter</code>.
     */
    public static void writeStatistics(Diallelic m, double[] statistics, 
            PrintWriter out) {
        StringBuilder sb = new StringBuilder();
        sb.append(m.id());
        sb.append(space);
        sb.append(m.refAllele());
        sb.append(space);
        for (int j=0; j<statistics.length; ++j) {
            sb.append(statistics[j]);
            sb.append(space);
        }
        double max = PermTester.max(statistics);
        if (max==Double.NEGATIVE_INFINITY) {
            max = Double.NaN;
        }
        sb.append(max);        
        out.println(sb.toString());
    }    
}    

