/*
 * PermTester.java
 *
 * Created on 26 July 2006, 19:24
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import java.util.*;
import java.io.*;
import blbutil.*;

/**
 * Class <code>PermTester</code> performs association tests
 * of a trait with a diallelic marker and computes p-values, minimum p-values,
 * and adjusted minimum p-values (from a permutation test of randomized
 * trait data).
 */
public class PermTester {

    private static final String newLine = System.getProperty("line.separator");
    private final String[] beagleFiles;
    private final String missingCode;
    
    /**
     * Creates a new instance of PermTester corresponding to the 
     * specified Beagle input files and Beagle output DAG files.
     *
     * @param beagleFiles an array of filenames of files in Beagle input
     * file format.
     * @param missingCode a string denoting a missing allele.   If 
     * <code>missingCode==null</code> there is no missing data.
     */
    public PermTester(String[] beagleFiles, String missingCode) {
        this.beagleFiles = (String[]) beagleFiles.clone();
        this.missingCode = missingCode;
    }
    
    /**
     * Returns <code>true</code> if the specified string 
     * describes a valid test and <code>false</code> otherwise.
     *
     * @param tests a string that deterines the tests to be performed.
     *
     * @return <code>true</code> if the specified string 
     * describes a valid test and <code>false</code> otherwise.
     */
    public static boolean isValidTest(String tests) {
        return FlexTest.isValidTest(tests);       
    }
    
    /**
     * Performs a permutation test using the specified trait file.
     * 
     * @param traitFile a file in Beagle input format containing at least one
     *        trait status line.  Only the first trait status line will be used.
     * @param strataFile a file in BEAGLE input format containing at least 
     *        one strata line (i.e. line whose first field is "S").  
     *        Only the first stata line will be used.
     * @param outPrefix a prefix used for output file names.
     * @param log a <tests>PrintWriter</tests> for writing a summary of the 
     *        permutation test.
     * @param nPerms the number of permutations of the trait data that will 
     *        be used.
     * @param seed the seed for the random number generator
     * @param isAllelicData a boolean that is <tests>true</tests> if the data 
     *        is not diplotype data (paired alleles), and <tests>false</tests> 
     *        otherwise.
     * @param tests a string that determines the tests to be performed.
     * @param maxSavedRanks the number of largest test statistics to save for
     *        for each permutation of the trait status.
     * @param threshold the first-stage test-statistic threshold for a 
     *        two-stage design.
     * @throws NullPointerException if <code>traitFile==null || 
     *         outPrefix==null || log== null || tests==null</code>.
     * @throws IllegalArgumentException if <code>nPerms < 0 || 
     *         maxSavedRank < 1 || threshold < 0.0</code>, if  
     *         <code>isAllelicData && !tests.equalsIgnoreCase("a")</code>, 
     *         or if <code>PermTester.isValidTest(tests)==false</code>.
     */
    public void test(String traitFile, String strataFile, String outPrefix, 
            PrintWriter log, int nPerms, long seed, boolean isAllelicData, 
            String tests, int maxSavedRanks, double threshold) {
        if (outPrefix==null) {
            throw new NullPointerException("outPrefix==null");
        }
        if (log==null) {
            throw new NullPointerException("log==null");
        }
        boolean[] origTrait = readTraitStatus(traitFile);
        int[] strata = readStrata(strataFile, this.missingCode);
        if (strata==null) {
            strata = new int[origTrait.length];
        }
        Test test = createTest(origTrait, strata, tests, isAllelicData);
        NullDistribution nullDistribution = new NullDistribution(origTrait, 
                strata, nPerms, maxSavedRanks, seed, isAllelicData, test, 
                threshold);
        File tmpStatFile = Utilities.tempFile("PermTester");
        computeTestStatistics(tmpStatFile, nullDistribution, strata); 
        nullDistribution.printNullDistribution(new File(outPrefix + ".null"));
        if (threshold <= 0.0) {
            PValueWriter pvw = new PValueWriter(tmpStatFile, outPrefix + ".pval", 
                        log, nullDistribution);
        }
        tmpStatFile.delete();
    }   
   
    /*
     * Returns a <code>permtester.Test</code> for the specified tests
     * and the case and control counts determined by the specified trait.
     *
     * @param trait a boolean array whose <code>true</code> entries
     * denote case alleles and whose <code>false</code> entries denote
     * control alleles.
     * @param strata an integer array whose k-th element gives the 
     * population stratum of the k-th allele.  All array elements 
     * must be non-negative.
     * @param tests a string whose characters determine the tests 
     * performed by the returned <code>permtester.Test</code> instance.
     * @param isAllelicData a boolean that is <tests>true</tests> if the data 
     * is not diplotype data (paired alleles), and <tests>false</tests> otherwise.
     * 
     * @return a <code>permtester.Test</code> for the specified tests
     * and the case and control counts determined by the specified trait.
     * @throws IllegalArgumentException if 
     * <code>isAllelicData && !code.equalsIgnoreCase("a")</code> or 
     * <code>trait.length != strata.length</code>, or if any 
     * element of the population strata array is negative.
     */
    private Test createTest(boolean[] trait, int[] strata, String tests, 
            boolean isAllelicData) {
        if (trait.length != strata.length) {
            String s = "number of alleles in affection status line ";
            s += "and population strata do not match";
            throw new IllegalArgumentException(s);
        }
        if (isAllelicData) {
            if (!tests.equalsIgnoreCase("a")) {
                String s = "Allelic data: tests = " + tests + " != " + "a";
                throw new IllegalArgumentException(s);
            }
        }
        else {
            checkDiplotypes(trait, strata);
        }
        int max = max(strata);
        if (max<0) {
            String s = "population strata cannot be negative: ";
            throw new IllegalArgumentException(s + Arrays.toString(strata));
        }
        int[] cases = new int[max+1];
        int[] controls = new int[max+1];
        for (int j=0; j<strata.length; ++j) {
            if (trait[j]) {
                ++cases[strata[j]];
            }
            else {
                ++controls[strata[j]];
            }
        }
        return new FlexTest(tests, cases, controls);
    }
    
    /*
     * Checks that the boolean trait array has even length, and that each 
     * <code>trait[2*j] == trait[2*j+1] && strata[2*j] == strata[2*j+1]</code> 
     * for each <code>0 &le; j < trait.length/2</code>.  The contract for
     * this method assumes that <code>trait.length == strata.length</code>.
     *
     * @param trait a boolean array giving binary trait status for each
     * allele in the sample.
     * @param strata an integer array giving population stratum for each
     * allele in the sample.
     *
     * @throws IllegalArgumentException if the specified trait array 
     * if <code>(trait.length & 1) != 0</code> or if
     * <code>trait[2*j] != trait[2*j+1] && strata[2*j] != strata[2*j+1]</code> 
     * for each <code>0 &le; j < trait.length/2</code>.
     */
    private void checkDiplotypes(boolean[] trait, int[] strata) {
        assert (trait.length == strata.length);
        if ((trait.length & 1) != 0) {
            StringBuilder sb = new StringBuilder();
            sb.append("odd number of trait alleles: ");
            sb.append(trait.length);
            sb.append(newLine);
            sb.append("Use \"diplotypes=false\" option with non-diploid data");
            throw new IllegalArgumentException(sb.toString());
        }
        for (int j=0; j<trait.length; j+=2) {
            if (trait[j] != trait[j+1]) {
                StringBuilder sb = new StringBuilder();
                sb.append("Alleles for individual ");
                sb.append(((j/2)+1));
                sb.append(" do not have identical affection status ");
                sb.append(newLine);
                sb.append("Use \"diplotypes=false\" option with non-diploid data");
                throw new IllegalArgumentException(sb.toString());
            }
            if (strata[j] != strata[j+1]) {
                StringBuilder sb = new StringBuilder();
                sb.append("Alleles for individual ");
                sb.append(((j/2)+1));
                sb.append(" do not have identical population stratum ");
                sb.append(newLine);
                sb.append("Use \"diplotypes=false\" option with non-diploid data");
                throw new IllegalArgumentException(sb.toString());                
            }
        }  
    }
    
    /**
     * Reads the first affection status line of the specified file
     * in Beagle input format and returns a boolean array with 
     * <code>true</code> entries
     * corresponding to affected alleles and <code>false</code>
     * entries corresponding to unaffected alleles.  Alleles
     * are not permitted to have an unknown affection status.
     *
     * @traitFile a file in Beagle input file format that contains at 
     * least one line specifying affection status.
     *
     * @throws IllegalArgumentException if the specified trait file
     * does contain any affection status data or is not in Beagle
     * input file format.
     */
    public static boolean[] readTraitStatus(String traitFile) {
        LineIterator it = new LineIterator(traitFile);
        boolean[] trait = null;
        while (it.hasNext() && trait==null) {
            String line = it.next().trim();
            String[] fields = StringUtil.getFields(line, 3);
            if (fields.length>0 && fields[0].equals("A")) {
                if (fields.length==2) {
                    String s = "missing trait data in file " + traitFile;
                    throw new IllegalArgumentException(s);
                }
                else {
                    fields = StringUtil.getFields(fields[2]);
                    trait = stringToBoolean(line, fields, "2", "1");
                }
            }
        }
        if (!it.hasNext() && trait==null) {
            String s = "Missing trait data in file " + traitFile;
            throw new IllegalArgumentException(s);          
        }   
        it.close();
        return trait;
    }

    /**
     * Reads the first strata line of the specified strata file
     * in Beagle input format and returns an <code>int</code> 
     * array whose k-th element is the stratum of
     * the k-th allele or else returns <code>null</code> if the 
     * specified strata file is the empty string (""). 
     * Alleles are not permitted to have unknown population stratum.
     * 
     * @strataFile a file in Beagle input file format that contains at 
     * least one strata line specifying population strata.
     * @param missingCode string representing missing data.  Alleles
     * are not permitted to have missing population stratum data.
     *
     * @return an <code>int</code> array whose k-th element is the 
     * stratum of each allele of the k-th allele or else returns 
     * <code>null</code> if the specified strata file is the empty 
     * string ("").
     *
     * @throws IllegalArgumentException if the specified strata file
     * does not contain a strata line or is not in Beagle
     * input file format.
     */
    public static int[] readStrata(String strataFile, String missingCode) {
        if (strataFile.equals("")) {
            return null;
        }
        LineIterator it = new LineIterator(strataFile);
        int[] strata = null;
        while (it.hasNext() && strata==null) {
            String line = it.next().trim();
            String[] fields = StringUtil.getFields(line, 3);
            if (fields.length>0 && fields[0].equals("S")) {
                if (fields.length==2) {
                    String s = "missing strata data in file " + strataFile;
                    throw new IllegalArgumentException(s);
                }
                MarkerLineInterface a = new MarkerLine(fields[1], fields[2], null);
                for (int j=0, n=a.nAlleleCodes(); j<n; ++j) {
                    if (missingCode.equals(a.allele(j))) {
                        String s = "Missing allele code cannot be used for ";
                        s += " population strata: " + missingCode;
                        throw new IllegalArgumentException(s);
                    }
                }
                strata = a.data();
            }
        }
        if (!it.hasNext() && strata==null) {
            String s = "Missing strata data in file " + strataFile;
            throw new IllegalArgumentException(s);          
        }   
        it.close();
        return strata;
    }    
    
    /**
     * Iterates through diallelic markers, prints the test statistics
     * from association testing with the unpermuted trait status to the
     * specified print writer, and computes and saves the distribution
     * of test statistics obtained from the permuted trait status.
     * 
     * @param tmpFile a file to which to write test statistics obtained using
     * the unpermuted trait status.  Data will only be written to 
     * <code>tmpFile</code> if <code>this.threshold() <= 0.0</code>.
     * @param nullDistibution a <code>NullDistribution</code> object to
     * compute and save the distribution of test statistics obtained 
     * from the permuted trait status.
     * @param strata an <code>int</code> array whose k-th element is the 
     * popoulation stratum of the k-th allele.
     */
    private void computeTestStatistics(File tmpFile, 
            NullDistribution nullDistribution, int[] strata) {
        int cnt = 0;
        boolean writePValues = nullDistribution.threshold() <= 0.0;
        boolean[] origTrait = nullDistribution.trait();
        Test test = nullDistribution.test();
        PrintWriter out = Utilities.createPrintWriter(tmpFile);
        FileIterator<Diallelic> it = getMarkerIterator(strata);
        while (it.hasNext()) {
            Diallelic m = it.next();
            nullDistribution.permTest(m);
            if (writePValues) {
                double[] pVals = test.pValues(origTrait, m);
                PValueWriter.writeStatistics(m, pVals, out);
            }
            if ((++cnt % 50000)==0) {
                System.out.println("Tested " + cnt + " markers");                    
            }
        }
        it.close();
        out.close();
        System.out.println("Tested " + cnt + " markers");                    
    }
    
    /*
     * Returns a diallelic marker iterator that iterates through 
     * all files in the <code>this.beagleFiles</code> array.
     *
     * @param strata an <code>int</code> array whose k-th element is the 
     * popoulation stratum of the k-th allele.
     *
     * @return a diallelic marker iterator that iterates through 
     * all files in the <code>this.beagleFiles</code> array.
     */
    private FileIterator<Diallelic> getMarkerIterator(final int[] strata) {
        return new AbstractMultiFileIterator<Diallelic>(PermTester.this.beagleFiles) {           
            protected InputIterator createIterator(String filename) {
                return new InputIterator(filename, PermTester.this.missingCode, 
                        strata);            
            }
        } ;        
    }
    
    /**
     * Returns the maximum value from the specified array.  If all 
     * elements of the specified array equal <code>Double.NaN</code> 
     * or if the specified array has length 0, 
     * then <code>-Double.POSITIVE_INFINITY</code> is returned.
     *
     * @param da a double array.
     * @return the maximum value from the specified array. 
     *
     * @throws NullPointerException if <code>da==null</code>.
     */
    public static double max(double[] da) {
        double max = Double.NEGATIVE_INFINITY;
        for (double d : da) {
            if (d > max) {
                max = d;
            }
        }
        return max;
    }
    
    /**
     * Returns the maximum value from the specified array.  
     *
     * @param ia an <code>int</code> array.
     * @return the maximum value from the specified array. 
     *
     * @throws NullPointerException if <code>da==null</code>.
     * @throws IllegalArgumentException if <code>ia.length==0</code>.
     */
    public static int max(int[] ia) {
        if (ia.length==0) {
            throw new IllegalArgumentException("array length==0");
        }
        int max = Integer.MIN_VALUE;
        for (int i : ia) {
            if (i > max) {
                max = i;
            }
        }
        return max;
    }
    
    /*
     * Transforms the specified <code>String</code> array to a boolean
     * array by converting the specified <code>t</code> string to 
     * <code>true</code> and the specified <code>f</code> string to 
     * false.
     *
     * @param line line of BEAGLE input file to be printed if 
     * exception is thrown.
     * @param sa array of string codes to be converted to boolean
     * array.
     * @param t string code to be converted to <code>true</code>.
     * @param f string code to be converted to <code>false</code>.
     * 
     * @return a boolean array constructed by converting the specified 
     * <code>t</code> string to <code>true</code> and the specified
     * <code>f</code> string to false.
     * 
     * @throws IllegalArgumentException if any entry in the specified 
     * <code>String</code> array is not equal to either <code>t</code>
     * or <code>f</code>
     * @throws NullPointerException if 
     * <code>sa==null || t==null || f==null</code>.
     */
    private static boolean[] stringToBoolean(String line, String[] sa, String t, 
            String f) {
        boolean[] ba = new boolean[sa.length];
        for (int j=0; j<sa.length; ++j) {
            String s = sa[j];
            if (s.equals(t)) {
                ba[j]=true;
            }
            else if (!s.equals(f)) {
                String msg = "Error in affection status line.  Affection status must be ";
                msg += t + " or " + f + ", and cannot be " + s + "." + newLine + line;
                throw new IllegalArgumentException(msg);
            }
        }
        return ba;
    }
}