/*
 * PrestoMain.java
 *
 * Created on 1 June 2006, 13:28
 *
 */

package permtester;

/**
 * @author Brian L Browning
 */

import blbutil.validator.*;
import java.io.*;
import java.util.*;
import blbutil.*;

 /*
  * See <code>err()</code> and <code>validators()</code> for 
  * for restrictions on <code>args parameters</code>.
  */
public class PrestoMain {
    
    private static final String newLine = System.getProperty("line.separator");
    private static final String version = "Presto 1.0.1";
    private static final List<Validator> validators = validators();

    private String err = null;
    private PrintWriter log = null;
    private Map<String, String> argsMap = null;
    private String outPrefix = null;
    private String[] markerFiles = null;
     
    /*
     * Returns list of validators for checking command line arguments.
     */
    private static List<Validator> validators() {
        List<Validator> list = new ArrayList<Validator>();
        list.add(new StringValidator("out", "<output file prefix>", "", true, false, null));
        list.add(new StringValidator("trait", "<trait file>", "", true, false, null));
        list.add(new StringValidator("strata", "<strata file>", "", false, false, null));
        list.add(new StringValidator("missing", "<missing allele code>", "", true, false, null));
        list.add(new CharStringValidator("test", "<association tests>", "t", 
                false, false, new char[] {'a', 'r', 'd', 'o', 't'}));
        list.add(new LongValidator("seed", "<random seed>", "-99999", false, false, 
                Long.MIN_VALUE, Long.MAX_VALUE ));
        list.add(new IntValidator("nperms", "<number of permutations>", "1000", 
                false, false, 0, Integer.MAX_VALUE));
        list.add(new IntValidator("topranks", "<number of order statistics>", 
                "1", false, false, 1, Integer.MAX_VALUE));
        list.add(new StringValidator("diplotypes", "<true/false>", "true", false, 
                false, new String[] {"true", "false"} ));
        list.add(new DoubleValidator("threshold", "<threshold for 2nd stage>", 
                "0.0", false, false, 0, Double.MAX_VALUE));
        return list;
    }
    
    /*
     * Sets the <code>this.err</code> field with the error message to be 
     * printed if arguments are not specified correctly.  Advanced arguments
     * are printed if they are used in the specified command line arguments.
     *
     * @param args command line arguments
     */   
    private void setErr(String[] args) {
        StringBuilder sb = new StringBuilder();
        sb.append(version);
        sb.append(newLine);
        sb.append(newLine);
        sb.append("Command line syntax: java -jar presto.jar [arguments] [files]");
        sb.append(newLine);
        sb.append(newLine);
        sb.append("[arguments] are ...");
        sb.append(newLine);
        for (Validator v : validators) {
            if (!v.isAdvanced()) {
                sb.append(getHelpMessage(v));
                sb.append(newLine);
            }
            else {
                String key = v.key();
                for (String s : args) {
                    if (s.startsWith(key)) {
                        sb.append(getHelpMessage(v));
                        sb.append(newLine);
                    }
                }
            }
        }
        sb.append(newLine);
        sb.append("[files] is a space-separated list of files in BEAGLE format");
        sb.append(newLine);
        sb.append(newLine);
        sb.append(testHelp());
        err = sb.toString();
    }
    
    /*
     * Returns a string message describing how to specify the association
     * tests to be performed.
     */    
    private static String testHelp() {
        StringBuilder sb = new StringBuilder();
        sb.append("<association tests> selects the tests that will be performed, and is one or");
        sb.append(newLine);
        sb.append("more characters from the string \"tardo\". Example: \"test=t\" or \"test=ard\" where");
        sb.append(newLine);
        sb.append(newLine);
        sb.append("  t = allelic trend test (default allelic test)");
        sb.append(newLine);
        sb.append("  a = allelic test");
        sb.append(newLine);
        sb.append("  r = recessive test     (groups major allele homozygotes and heterozygotes)");
        sb.append(newLine);
        sb.append("  d = dominant test      (groups minor allele homozygotes and heterozygotes)");
        sb.append(newLine);
        sb.append("  o = overdominant test  (groups minor and major allele homozygotes)");
        sb.append(newLine);
        return sb.toString();
    }
    
    /*
     * Returns a help message corresponding to the specified validator.
     */
    private static String getHelpMessage(Validator v) {
        StringBuilder sb = new StringBuilder();
        sb.append("  ");
        sb.append(String.format("%-40s", v.toString()));
        if (v.isRequired()) {
            sb.append("  (required)");
        }
        else {
            sb.append("  (optional");
            if (v.defaultValue()!="") {
                sb.append(": default=");
                sb.append(v.defaultValue());
            }
            sb.append(")");
        }
        return sb.toString();
    }
    

    
    /**
     * Prints the specified command line to the specified PrintWriter
     *
     * @param args command line arguments.
     * @param out a print writer.
     */
    private void printCommandLine(String[] args, PrintWriter out) {
        if (args.length==0) {
            printErrorAndExit("");
        }
        else {
            out.println("Command line: java -Xmx600m -jar presto.jar ");
            for (int j=0; j<args.length; ++j) {
                out.print("  ");
                out.println(args[j]);
            }
            out.println();
        }        
    }
    
    /**
     * Creates a map that maps parameters to values from the 
     * specified command line arguments and stores map in the 
     * private <code>this.argsMap</code> field.  Prints error and exits 
     * if any entry in <code>argsMap</code> fails to validate or 
     * if there are any keys in <code>argsMap</code>
     * that do not correspond to any <code>Validator</code>.
     *
     * @param args the command line arguments.
     */
    private void readArgsMap(String[] args) {
        argsMap = Utilities.argsToMap(args, '=');
        for (Validator v : validators) {
            try {
                v.validate(argsMap);
            }
            catch (IllegalArgumentException e) {
                printErrorAndExit(e.toString());
            }
        } 
        // xxx check for unexpected arguments
        Collection<String> c = new HashSet<String>(argsMap.keySet());
        for (Validator v : validators) {
            c.remove(v.key());
        }
        if (c.size() > 0) {
            System.out.print("Error: unrecognized arguments:");
            for (String key : c) {
                System.out.println("  " + key + "=" + argsMap.get(key));
            }
            printErrorAndExit("");
        }
        boolean isAllelicData = !Boolean.parseBoolean(argsMap.get("diplotypes"));
        if (isAllelicData && !argsMap.get("test").equalsIgnoreCase("a")) {
            String s = "Error: only allelic test (test=a) can be used ";
            s += "when diplotypes=false";
            printErrorAndExit(s + newLine);
        }
    }

    /**
     * Stores all file names in the specified command line arguments 
     * as a string array in the private <code>this.markerFiles</code> field.
     * A string that is a command line argument is considered to be a filename 
     * if the string does not contain the '=' character in the argument.
     *
     * @param args command line arguments
     * @throws IllegalArgumentException if any files cannot be read.
     */
    private void getMarkerFiles(String[] args) {
        List<String> argsList = Utilities.argsToList(args, '=');
        if (argsList.size()==0) {
            printErrorAndExit("Error: no filenames specified");
        }
        for (String s : argsList) {
            File f = new File(s);
            if (!f.canRead()) {
                printErrorAndExit("Cannot read file " + s);
            }
        }
        this.markerFiles = argsList.toArray(new String[0]);
    }
    
    /*
     * Constructs a <code>PhaseMain</code> instance that runs 
     * the Presto program using the specified arguments.
     * Private constructor prevents public instantiation
     *
     * @param args command line arguments
     */ 
    private PrestoMain(String[] args) {
        setErr(args);
        printCommandLine(args, new PrintWriter(System.out, true));
        readArgsMap(args);
        getMarkerFiles(args); 
        this.log = Utilities.createPrintWriter(argsMap.get("out") + ".log");
        log.println(err);
        printCommandLine(args, log);
        testFiles();
        String s = version + " finished";
        System.out.println(s);
        log.println(s);
        log.close();
    }    

    /*
     * Tests for association of trait and markers using permutation to 
     * adjust for multiple testing.
     */
    private void testFiles() {
        String outPrefix = argsMap.get("out");
        String traitFile = argsMap.get("trait");
        String strataFile = argsMap.get("strata");
        int nPerms = Integer.parseInt(argsMap.get("nperms"));
        long seed = Long.parseLong(argsMap.get("seed"));
        String tests = argsMap.get("test");
        boolean isAllelicData = !Boolean.parseBoolean(argsMap.get("diplotypes"));
        String missingCode = argsMap.get("missing");
        if (missingCode.equals("")) {
            missingCode = null;
        }
        int nLargest = Integer.parseInt(argsMap.get("topranks"));
        double threshold = Double.parseDouble(argsMap.get("threshold"));
        
        long before = System.currentTimeMillis();
        PermTester pt = new PermTester(markerFiles, missingCode);        
        pt.test(traitFile, strataFile, outPrefix, log, nPerms, seed, 
                isAllelicData, tests, nLargest, threshold);
        String s = newLine + "Running time: " + 
                Utilities.elapsedTime(System.currentTimeMillis() - before);
        log.println(s);
        System.out.println(s);
    }    

    /*
     * Prints the <code>this.err</code> field followed by the 
     * specified prefix to standard out and exists the java virtual machine.
     *
     * @param message a message to be written to standard out.
     */
    private void printErrorAndExit(String message) {
        System.out.println(err);
        System.out.println(message);
        System.out.flush();
        System.exit(0);       
    }
    
    /**
     * Entry point to Presto program.  See documentation for 
     * information on program arguments.
     */
    public static void main(String[] args) {
        PrestoMain main = new PrestoMain(args);
    }    
}
